from cobra import Reaction, Metabolite
from typing import Optional


class Constraint(Metabolite):
    """
    Class for information about a Constraint in a protein Sector.

    Constraint is a class for holding information similar to
    a metabolite in a cobra.Reaction object.

    Parameters:
        id (str): The identifier to associate with the constraint.
        name (str): A human-readable name.
    """

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        id: Optional[str] = None,
        formula: Optional[str] = None,
        name: Optional[str] = "",
        charge: Optional[float] = None,
        compartment: Optional[str] = "cytosol",
    ) -> None:
        super().__init__(
            id=id,
            formula=formula,
            name=name,
            charge=charge,
            compartment=compartment,
        )
        self.annotation = {"type": "Constraint"}



"""Classes related to enzymes:
 - Enzyme: Constraints relating enzymes to reactions. Including upper and lower bound enzyme constraints
 - EnzymeVariable: Variable related to an enzyme. The value of this variable represent the concentration.
"""
import cobra.core
import cobra
from cobra import Reaction, Object, Gene, DictList
from cobra.exceptions import OptimizationError
from cobra.util.solver import check_solver_status
from copy import copy, deepcopy

import optlang
from optlang.symbolics import Zero

from .CatalyticEvent import CatalyticEvent
from typing import Dict, Union, Optional
from cobra import DictList
from warnings import warn


def _change_catalytic_event_list_to_dictlist_after_unpickling(self):
    # return lists back to dictlist after unpickling
    if not isinstance(self.catalytic_events,DictList):
        self.enzymes = DictList(self.catalytic_events)


class Enzyme(Object):
    """Upper level Enzyme object containing information about the enzyme and links to the EnzymeVariables for each reaction the enzyme catalyzes.

    Args:
        - id (str): Identifier for the enzyme (e.g., Uniprot ID).
        - rxn2kcat (Dict): Dictionary with reaction ID, kcat value pairs for the forward (f) and backward (b) reaction,
            e.g. `{'PGI': {'f': 30, 'b': 0.1}}`
        - upper_bound (float): Upper bound for the enzyme variable (default 1000.0).
        - lower_bound (float): Lower bound for the enzyme variable (default 0).
        - name (str): Name of the enzyme (default None).
        - molmass (float): Molar mass of the enzyme (default 3.947778784340140e04).

    Notes:
        - This class is used to generate enzyme instances from kcat values and contains information about the forward as well as the backward catalysis.
        - The enzyme is linked to individual cobra.Reaction variables with CatalyticEvent objects.
        - There are two scenarios:
            - Promiscuous enzymes: a single enzyme can catalyze multiple reactions.
            - Other: a single enzyme catalyzes a single reaction.
    """

    # constant parameters
    DEFAULT_ENZYME_MOL_MASS = 3.947778784340140e04  # mean enzymes mass E.coli [g/mol]

    def __init__(
        self,
        id: str,
        rxn2kcat: Dict,
        rxn2pr: Dict = {},
        genes: list = [],
        upper_bound: Union[int, float] = 1000.0,
        lower_bound: Union[int, float] = 0,
        name: Optional[str] = None,
        molmass: Union[int, float] = DEFAULT_ENZYME_MOL_MASS,
    ):
        self.rxn2kcat = rxn2kcat
        self.rxn2pr = rxn2pr
        self.molmass = molmass
        self.id = id  # use e.g. Uniprot ID
        self.name = name
        self.upper_bound = upper_bound
        self.lower_bound = lower_bound

        # create enzyme variable associated with theis enzyme
        self.enzyme_variable = None
        self.create_enzyme_variable()

        self.catalytic_event_id = (
            "CE_" + "{0}"
        )  # generic template for the catalytic events IDs
        # initialize CatalyticEvent interfaces for each associated reaction (promiscuity)
        self.catalytic_events = DictList()
        for rxn_id, kcats in rxn2kcat.items():
            self.create_catalytic_event(rxn_id, kcats)

        self._constraints = (
            {}
        )  # dict with constraint_id:optlang.Constraint, key:value pairs.
        self._model = None
        self.enzyme_complex = []  # is the enzyme in a complex?
        self.genes = genes
        self.transcripts = DictList()
        self.annotation = {
            "type": "Constraint"
        }  # you can add an annotation for an enzyme
    @property
    def kcat_values(self):
        """Returns a dictionary with kcat values for each associated reaction.

        Returns:
            dict: A dictionary containing kcat values for associated reactions.
        """
        return self.get_kcat_values()

    @property
    def concentration(
        self, units: str = "mmol/gDW", return_units: bool = False
    ) -> float:
        """Returns the enzyme's total concentration considering any associated reactions.

        Args:
            units (str, optional): Units in which the concentration is calculated (default is 'mmol/gDW'), other option is 'g/gDW'.
            return_units (bool, optional): Determines whether the units should be returned as well.

        Returns:
            float: Enzyme concentration as a float.
        """

        # sum up concentrations (aka fluxes) of all enzyme objects
        concentration = self.enzyme_variable.concentration
        if units == 'g/gDW':
            #converting mmol to grams of protein:
            # [g] = [mmol]* 1e-3 [mol/mmol] * MW[g/mol]
            concentration = concentration * 1e-3 * self.molmass
        if return_units:
            return concentration, units
        return concentration

    @concentration.setter
    def concentration(self, conc):
        self.enzyme_variable.concentration = conc

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model):
        self._model = model

    def set_forward_concentration(self, conc):
        self.enzyme_variable.set_forward_concentration(conc)

    def set_reverse_concentration(self, conc):
        self.enzyme_variable.set_reverse_concentration(conc)

    def create_constraint(self, extension: str = None):
        if extension is None:
            extension = ""
        else:
            extension = "_" + extension
        # return cobra.core.Metabolite(id = f'EC_{self.id}{extension}', compartment='Enzymes')
        return optlang.Constraint(Zero, name=f"EC_{self.id}{extension}", lb=0, ub=0)

    def add_catalytic_event(self, ce: CatalyticEvent, kcats: Dict):
        """Adds a catalytic event associated with a reaction to an enzyme.

        Args:
            ce (PAModelpy.Variables.CatalyticEvent): A CatalyticEvent object to which the enzyme should be added.
            kcats (dict): A dictionary containing direction and kcat key-value pairs.

        Returns:
            NoneType: None
        """
        _change_catalytic_event_list_to_dictlist_after_unpickling(self)
        self.catalytic_events += [ce]
        self.enzyme_variable.add_catalytic_events([ce], [kcats])

    def add_genes(self, gene_list: list, gene_length:list, relation:str = 'OR') -> None:
        """
            Add genes to the enzyme and the model related to the enzyme if applicable

            Args:
                gene_list (list): A list of gene identifiers to be added.
                gene_length (list): A list of lengths corresponding to each gene.
                relation (str, optional): The relationship between genes in gene_list.
                    Defaults to 'OR'. Possible values: 'OR' or 'AND'.

            Raises:
                ValueError: If an invalid relation is provided.

            Note:
                If relation is 'OR', each gene in gene_list will be treated as coding for an individual isozyme
                If relation is 'AND', all genes in gene_list will be treated as coding for peptides in an enzyme complex

            """
        # check/correct type of arguments
        if isinstance(gene_list, str): gene_list = [gene_list]
        if not isinstance(gene_length, list): gene_length = [gene_length]

        if relation == 'OR':
            genes_to_add = [[Gene(gene)] for gene in gene_list]
        elif relation == 'AND':
            genes_to_add = [[Gene(gene) for gene in gene_list]]
        else:
            raise ValueError("Invalid relation. Supported values are 'OR' or 'AND'.")
        self.genes += genes_to_add

        if self._model is not None:
            self._model.add_genes(genes = genes_to_add, enzymes = [self], gene_lengths=gene_length)


    def create_catalytic_event(self, rxn_id: str, kcats: Dict):
        """Creates enzyme variables that link to reactions.

        Args:
            rxn_id (str): ID of the associated reaction in the model.
            kcats (Dict): A dictionary containing kcat values for the forward and backward reaction.

        Returns:
            Variables.CatalyticEvent: Enzyme variable object of type Variables.CatalyticEvent.
        """

        # create unique enzyme object name and id
        catalytic_event_id = self.catalytic_event_id.format(rxn_id)
        if self.name is not None:
            enzyme_object_name = rxn_id + "_" + self.name
        else:
            enzyme_object_name = rxn_id + "_" + self.id

        # create enzymatic reaction object inherited from cobra.Reaction
        catalytic_event = CatalyticEvent(
            id=catalytic_event_id,
            rxn_id=rxn_id,
            kcats2enzymes={self: kcats},
            name=enzyme_object_name,
        )

        self.add_catalytic_event(catalytic_event, kcats)

    def create_enzyme_variable(self):
        """Creates enzyme variables that link  enzyme to reactions."""
        # create enzymatic reaction object inherited from cobra.Reaction
        enzyme_variable = EnzymeVariable(
            id=self.id,
            kcats2rxns=self.rxn2kcat,
            upper_bound=self.upper_bound,
        )

        self.enzyme_variable = enzyme_variable

    def change_kcat_values(self, rxn2kcat: Dict):
        """Changes the kcat values for the enzyme and updates the enzyme variable (enzymatic reaction) accordingly.

        Args:
            rxn2kcat (Dict): A dictionary with reaction ID, kcat value pairs for the forward (f) and backward (b) reaction,
                e.g. `{'PGI': {'f': 30, 'b': 0.1}}`
        """
        _change_catalytic_event_list_to_dictlist_after_unpickling(self)

        # update the enzyme variables
        for rxn_id, kcats in rxn2kcat.items():
            catalytic_event_id = self.catalytic_event_id.format(rxn_id)
            # change rxn2kcat dictionary
            self.rxn2kcat[rxn_id] = kcats
            # is there already a link between enzyme and reaction?
            if catalytic_event_id not in self.catalytic_events:
                warn(f"Reaction {rxn_id} is not associated with enzyme {self.id}. Skip")
            else:
                # change kcat values of existing enzyme variable
                self.catalytic_events.get_by_id(catalytic_event_id).change_kcat_values(
                    {self.id: kcats}
                )

    def get_kcat_values(self, rxn_ids: Union[str, list] = None) -> Dict:
        """Returns the kcat values for a specific enzyme and all enzyme-associated reactions.

        Args:
            rxn_ids (Union[str, list], optional): ID of the reactions for which the kcat values should be returned. It can be a single reaction ID (str) or a list of reaction IDs.

        Returns:
            Dict: A dictionary containing kcat values for the forward (f) and backward (b) reactions.
        """

        if isinstance(rxn_ids, str):
            rxn_ids = [rxn_ids]

        rxn2kcat = {}
        if rxn_ids is None:
            # return all kcat values
            rxn2kcat = self.rxn2kcat
        else:
            # return reaction specific kcat values
            for rxn_id in rxn_ids:
                catalytic_event_id = self.catalytic_event_id.format(rxn_id)
                if catalytic_event_id not in self.catalytic_events:
                    warn("No reaction {0} found".format(rxn_id))
                else:
                    # get kcat values
                    rxn2kcat[rxn_id] = self.rxn2kcat[catalytic_event_id+'_'+self.id]

        if len(rxn_ids) == 1:
            return rxn2kcat[rxn_id]

        return rxn2kcat

    def remove_catalytic_event(self, catalytic_event: Union[CatalyticEvent, str]):
        """Function to remove a catalytic event from an enzyme.

        Args:
            catalytic_event (Union[CatalyticEvent, str]): CatalyticEvent or str, catalytic event or identifier to remove.
        """
        _change_catalytic_event_list_to_dictlist_after_unpickling(self)
        if isinstance(catalytic_event, str):
            try:
                catalytic_event = self.catalytic_events.get_by_id(catalytic_event)
            except:
                print(
                    f"Catalytic event {catalytic_event} is not related to this enzyme and can thus not be removed!"
                )

        # remove the event from the DictList
        self.catalytic_events.remove(catalytic_event)


    def __copy__(self) -> "Enzyme":
        """Copy the enzyme variable.

        Returns:
            PAModelpy.Enzyme.Enzyme: A new enzyme that is a copy of the original enzyme.
        """

        cop = copy(super(Enzyme, self))
        return cop

    def __deepcopy__(self, memo: dict) -> "Enzyme":
        """Copy the enzyme variable with memo.

        Args:
            memo (dict): Automatically passed parameter.

        Returns:
            PAModelpy.Enzyme.Enzyme: A new enzyme that is a copy of the original enzyme with memo.
        """

        cop = deepcopy(super(Enzyme, self), memo)
        return cop


    def __getstate__(self):
        # Return the state to be pickled
        state = self.__dict__.copy()
        state['catalytic_events'] = list(self.catalytic_events)
        # Handle any non-serializable attributes here
        return state

    def __setstate__(self, state):
        # Restore state from the unpickled state
        self.__dict__.update(state)
        # Handle any attributes that require initialization or special handling here



class EnzymeComplex(Enzyme):
    """Upper-level EnzymeComplex object containing information about the enzymes in a complex
    and a link to the enzyme variables (CatalyticEvents) for each reaction the enzyme complex catalyzes.

    This class is used to generate enzyme instances from kcat values and contains
    information about the forward as well as the backward catalysis.

    Args:
        id (str): Identifier for the enzyme complex (e.g., Uniprot ID).
        enzymes (DictList of cobra.core.Enzyme): Enzyme objects associated with the enzyme complex.
        rxn2kcat (Dict): Dictionary with reaction ID, kcat value pairs for the forward (f) and backward (b) reaction,
            e.g. `{'PGI': {'f': 30, 'b': 0.1}}`
        upper_bound (float, optional): Upper bound for the enzyme variable (default 1000.0).
        name (str, optional): Name of the enzyme (default None).
        molmass (float, optional): Molar mass of the enzyme (default 3.947778784340140e04).
    """

    # constant parameters
    DEFAULT_ENZYME_MOL_MASS = 3.947778784340140e04  # mean enzymes mass E.coli [g/mol]

    def __init__(
        self,
        id: str,
        rxn2kcat: Dict,
            enzymes: DictList = DictList(),
        genes: list = [],
            upper_bound: Union[int, float] = 1000.0,
            name: Optional[str] = None,
            molmass: Union[int, float] = DEFAULT_ENZYME_MOL_MASS, ):
        super().__init__(
            id=id,
            rxn2kcat=rxn2kcat,
            genes=genes,
            upper_bound=upper_bound,
            name=name,
            molmass=molmass,
        )

        self.reactions = list(rxn2kcat.keys())
        self.enzymes = DictList()
        self.add_enzymes(enzymes)

    def add_enzymes(self, enzymes: DictList):
        for enzyme in enzymes:
            if enzyme not in self.enzymes:
                self.enzymes.append(enzyme)
                enzyme.enzyme_complex.append(self.id)
                self.molmass += enzyme.molmass
                #remove associated constraints from the model
                if self._model is not None:
                    if enzyme in self._model.enzymes:
                        for rxn in self.reactions:
                            self._model.remove_enzyme_reaction_association(enzyme, rxn)
                    else:
                        self._model.enzymes.append(enzyme)

    def __copy__(self) -> "EnzymeComplex":
        """Copy the enzyme complex.

        Returns:
            PAModelpy.Enzyme.EnzymeComplex: A new enzyme complex that is a copy of the original enzyme.
        """

        cop = copy(super(EnzymeComplex, self))
        return cop

    def __deepcopy__(self, memo: dict) -> "EnzymeComplex":
        """Copy the enzyme complex with memo.

        Args:
            memo (dict): Automatically passed parameter.

        Returns:
            PAModelpy.Enzyme.EnzymeComplex: A new enzyme complex that is a copy of the original enzyme with memo.
        """

        cop = deepcopy(super(EnzymeComplex, self), memo)
        return cop


    def __getstate__(self):
        # Return the state to be pickled
        state = self.__dict__.copy()
        state['catalytic_events'] = list(self.catalytic_events)

        # Handle any non-serializable attributes here
        return state

    def __setstate__(self, state):
        # Restore state from the unpickled state
        self.__dict__.update(state)
        # Handle any attributes that require initialization or special handling here



class EnzymeVariable(Reaction):
    """EnzymeVariable is a class for holding information regarding the variable representing an enzyme in the model.
    For each reaction, the enzyme variables are summarized in a CatalyticEvent.

    There are three different scenarios:
    - Enzyme complex: multiple enzymes together are associated with an EnzymeComplex object.
    - Isozymes: multiple enzymes independently associated with a single catalytic event.
    - Other: a single enzyme is associated with a single catalytic event.

    Args:
        kcats2rxns (Dict): A dictionary with reaction_id, kcat key, value pairs to connect the enzyme with the associated reaction.
            The kcat is another dictionary with `f` and `b` for the forward and backward reactions, respectively.
        id (str, optional): The identifier to associate with this enzyme (default None).
        name (str, optional): A human-readable name for the enzyme (default "").
        subsystem (str, optional): Subsystem where the enzyme is meant to function (default "").
        lower_bound (float): The lower flux bound (default 0.0).
        upper_bound (float, optional): The upper flux bound (default None).
        **kwargs: Additional keyword arguments are passed on to the parent class.
    """

    DEFAULT_ENZYME_MOL_MASS = 3.947778784340140e04  # mean enzymes mass E.coli [g/mol]

    def __init__(
        self,
        kcats2rxns: Dict,
        molmass: Union[int, float] = DEFAULT_ENZYME_MOL_MASS,
        id: Optional[str] = None,  # ID of enzymatic reaction,
        name: str = "",
        upper_bound: Optional[float] = None,
        **kwargs,
    ):
        super().__init__(
            id=id,
            name=name,
            subsystem="Enzymes",
            lower_bound=-upper_bound,
            upper_bound=upper_bound,
            **kwargs,
        )
        self.kcats = kcats2rxns
        self.molmass = molmass
        self.rxn_ids = [rxn for rxn in kcats2rxns.keys()]
        self.enzyme = None #store the enzyme associated to the enzyme variable
        self.catalytic_events = DictList() #store the interfaces to reactions related to this enzyme
        self.reactions = DictList()
        self.constraints = {}  # store IDs of constraint the enzyme is associated with
        self._model = None
        self.annotation = {"type": "Variable"}
        self.variables = dict()

    @property
    def kcat_values(self):
        """Returns a dictionary with kcat values and reactions.

        Returns:
            dict: A dictionary containing kcat values and their associated reactions.
        """
        return self.kcats

    @property
    def flux(self) -> float:
        """Get the flux value in the most recent solution.

        Flux is the primal value of the corresponding variable in the model.

        Returns:
            float: Flux, which is the primal value of the corresponding variable in the model.

        Raises:
            RuntimeError: If the underlying model was never optimized beforehand or the reaction is not part of a model.
            OptimizationError: If the solver status is anything other than 'optimal'.
            AssertionError: If the flux value is not within the bounds.

        Warnings:
            - Accessing reaction fluxes through a `Solution` object is the safer, preferred, and only guaranteed to be correct way.
            - Reaction flux is retrieved from the currently defined `self._model.solver`. The solver status is checked but there are no guarantees that the current solver state is the one you are looking for.
            - If you modify the underlying model after an optimization, you will retrieve the old optimization values.

        Examples:
            ```
            >>> from cobra.io import load_model
            >>> model = load_model("textbook")
            >>> solution = model.optimize()
            >>> model.variables.PFK.flux
            7.477381962160283
            >>> solution.fluxes.PFK
            7.4773819621602833
            ```
        """

        try:
            check_solver_status(self._model.solver.status)
            return self.forward_variable.primal + self.reverse_variable.primal
        except AttributeError:
            raise RuntimeError(f"reaction '{self.id}' is not part of a model")
        # Due to below all-catch, which sucks, need to reraise these.
        except (RuntimeError, OptimizationError) as err:
            raise err
        # Would love to catch CplexSolverError and GurobiError here.
        except Exception as err:
            raise OptimizationError(
                f"Likely no solution exists. Original solver message: {str(err)}."
            ) from err

    @property
    def concentration(self) -> float:
        """Get the enzyme concentration value of the most recent solution.

        The enzyme concentration equals the flux value.

        Returns:
            float: Enzyme concentration in mmol/gDW."""
        return self.flux*1e-6

    @concentration.setter
    def concentration(self, conc:Union[float,int]) -> None:
        """
        Sets the concentration of the enzyme by creating or updating a constraint
        that enforces the concentration to be equal to the sum of the forward and reverse
        variable primals.

        Args:
        conc : float, int
            The concentration value to be set for the enzyme. This value will be used
            as both the lower and upper bound for the constraint, effectively fixing the
            concentration to this value.

        Notes
        -----
        - If a concentration constraint for the enzyme does not already exist in the model,
          this function creates a new constraint named '<enzyme_id>_conc'.
        - The concentration constraint is defined as:
          concentration = forward_variable.primal + reverse_variable.primal
        - If the constraint already exists, the linear coefficients for the forward and reverse
          variables are updated to ensure the constraint remains valid.

        Raises
        ------
        ValueError
            If `conc` is not a valid numerical value.
        """

        # Ensure `conc` is a valid numerical value
        if not isinstance(conc, (int, float)):
            raise ValueError("The concentration must be a numerical value.")

        # Make the constraint: concentration = forward_variable + reverse_variable
        if self.id + '_conc' not in self._model.constraints.keys():
            concentration_constraint = self._model.problem.Constraint(
                Zero, name=self.id + '_conc', lb=conc, ub=conc)
            self._model.add_cons_vars(concentration_constraint)

        #units of enzyme should be *1e6 because of feasibility tolerance solver
        self._model.constraints[self.id + '_conc'].set_linear_coefficients({
            self.forward_variable: 1e-6,
            self.reverse_variable: 1e-6
        })
        self.enzyme._constraints[self.id+'_conc'] = self._model.constraints[self.id + '_conc']


    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model):
        # _change_catalytic_event_list_to_dictlist_after_unpickling(self)

        self._model = model
        # setting up the relations to the model
        # add enzyme instance
        enzyme_ids = [enz.id for enz in self._model.enzymes]
        if self.id not in enzyme_ids:
            enzyme = Enzyme(id =self.id,
                            rxn2kcat = self.kcat,
                            molmass=self.molmass
                            )
            self._model.add_enzymes([enzyme])
        self.enzyme = self._model.enzymes.get_by_id(self.id)
        self.constraints = self.enzyme._constraints
        self._model.enzyme_variables.append(self)

        # create new forward and reverse variables
        forward_variable = self._model.problem.Variable(
            name=self.id, ub=self.upper_bound, lb=0
        )
        reverse_variable = self._model.problem.Variable(
            name=self.reverse_id, ub=-self.lower_bound, lb=0
        )
        self.variables = {
            "forward_variable": forward_variable,
            "reverse_variable": reverse_variable,
        }
        self._model.add_cons_vars([forward_variable, reverse_variable])

        # add catalytic event and reaction instances
        for rxn_id in self.rxn_ids:
            if not rxn_id in self._model.reactions:
                # create new reaction and add to model
                rxn = cobra.Reaction(id=rxn_id)
                self._model.add_reactions([rxn])
            if not rxn_id in self.reactions:
                self.reactions.append(self._model.reactions.get_by_id(rxn_id))
            # create a catalytic event if it doesn't exist already
            if not f"CE_{rxn_id}" in self._model.catalytic_events:
                kcats2enzymes = {self.enzyme: self.kcats[rxn_id]}
                ce = CatalyticEvent(
                    id=f"CE_{rxn_id}", kcats2enzymes=kcats2enzymes, rxn_id=rxn_id
                )
                ce.model = model
                self._model.catalytic_events.append(ce)
            # if catalytic event exist, add the enzyme to it
            else:
                self._model.catalytic_events.get_by_id(f"CE_{rxn_id}").add_enzymes(
                    {self.enzyme: self.kcats[rxn_id]}
                )
            # if catalytic event is not related to the enzyme variable yet, add it.
            if f"CE_{rxn_id}" not in self.catalytic_events:
                self.catalytic_events.append(
                    self._model.catalytic_events.get_by_id(f"CE_{rxn_id}")
                )

    @property
    def forward_variable(self):
        if self._model is not None:
            return self._model.variables[self.id]
        else:
            return None

    @property
    def reverse_variable(self):
        if self._model is not None:
            return self._model.variables[self.reverse_id]
        else:
            return None

    def set_forward_concentration(self, conc: Union[float, int]) -> None:
        """
        Sets the concentration of the enzyme by creating or updating a constraint
        that enforces the concentration to be equal to the sum of only the forward
        variable primals. This forces a reaction to be active in the forward direction.
        It used the concentration setter functionality and subsequently sets the
        coefficient for the reverse variable in the constraint to 0.

        Args:
        conc : float, int
            The concentration value to be set for the enzyme. This value will be used
            as both the lower and upper bound for the constraint, effectively fixing the
            concentration to this value.

        Notes
        -----
        - If a concentration constraint for the enzyme does not already exist in the model,
          this function creates a new constraint named '<enzyme_id>_conc'.
        - The concentration constraint is defined as:
          concentration = forward_variable.primal

        Raises
        ------
        ValueError
            If `conc` is not a valid numerical value.
        """
        self.concentration = conc
        self._model.constraints[self.id + '_conc'].set_linear_coefficients({
            self.forward_variable: 1,
            self.reverse_variable: 0
        })

    def set_reverse_concentration(self, conc: Union[float, int]) -> None:
        """
        Sets the concentration of the enzyme by creating or updating a constraint
        that enforces the concentration to be equal to the sum of only the reverse
        variable primals. This forces a reaction to be active in the reverse direction.
        It used the concentration setter functionality and subsequently sets the
        coefficient for the forward variable in the constraint to 0.

        Args:
        conc : float, int
            The concentration value to be set for the enzyme. This value will be used
            as both the lower and upper bound for the constraint, effectively fixing the
            concentration to this value.

        Notes
        -----
        - If a concentration constraint for the enzyme does not already exist in the model,
          this function creates a new constraint named '<enzyme_id>_conc'.
        - The concentration constraint is defined as:
          concentration = reverse_variable.primal

        Raises
        ------
        ValueError
            If `conc` is not a valid numerical value.
        """
        self.concentration = conc
        self._model.constraints[self.id + '_conc'].set_linear_coefficients({
            self.forward_variable: 0,
            self.reverse_variable: 1
        })

    def add_catalytic_events(self, catalytic_events:list, kcats:list):
        """
        Adding a catalytic event to an enzyme variable

        Args:
            catalytic_events (list): A list of catalytic events to add.
            kcats (list): A list with dictionaries containing direction and kcat key-value pairs.
        """
        _change_catalytic_event_list_to_dictlist_after_unpickling(self)

        for i, ce in enumerate(catalytic_events):
            if ce not in self.catalytic_events:
                if not ce.enzyme_variables.has_id(self.id):
                    ce.enzyme_variables.append(self)
                self.catalytic_events.append(ce)
            if not self.reactions.has_id(ce.rxn_id):
                self.add_reactions({ce.rxn_id: kcats[i]})

    def add_reactions(self, reaction_kcat_dict: dict):
        """Add reactions to the enzyme variable and create bindings to the related model.
        If there are multiple reactions related to a single enzyme, this is an isozyme.

        Args:
            reaction_kcat_dict (dict): A nested dictionary with the reaction_id, kcat key, value pairs to connect the
                enzyme with the associated reaction. The kcat is another dictionary with `f` and `b` for the forward and
                backward reactions, respectively.
        """
        for reaction, kcat in reaction_kcat_dict.items():
            # check if the enzyme is already associated to the catalytic event
            try:
                self.reactions.get_by_id(reaction.id)
                warn(
                    f"Reaction {reaction.id} is already associated with the enzyme {self.id}. The enzyme variable will be updated"
                )
                self.change_kcat_values(kcat)
                return
            except:
                pass

            self.kcats[reaction] = kcat

            if self._model is None:
                continue

            # check if enzyme is in the model
            try:
                self._model.reactions.get_by_id(reaction)
            # if not: add the enzyme to the model
            except:
                rxn = Reaction(id=reaction)
                self._model.add_reactions([rxn])

            rxn = self._model.reactions.get_by_id(reaction)
            self.reactions.append(rxn)

            for direction in kcat.keys():
                if direction != "f" and direction != "b":
                    warn(
                        f"Invalid direction {direction} for kcat value for enzyme variable {self.id}! Skip!"
                    )
                    continue

            # add enzyme to catalytic event and the related variable
            for direction, kcatvalue in kcat.items():
                self._model._change_kcat_in_enzyme_constraint(rxn, self.id,
                                                              direction, kcatvalue)
                # add enzyme to the associated reaction with kinetic constants
                # and relate enzyme to the catalytic event
                if direction == 'f':
                    self.constraints[f'EC_{self.id}_{direction}'].set_linear_coefficients({
                        self.forward_variable: -1
                    })

                elif direction == 'b':
                    self.constraints[f'EC_{self.id}_{direction}'].set_linear_coefficients({
                        self.reverse_variable: -1
                    })


    def remove_catalytic_event(self, catalytic_event: Union[CatalyticEvent, str]):
        """Remove a catalytic event from an enzyme.

        Args:
            catalytic_event (Union[CatalyticEvent, str]): CatalyticEvent or str, catalytic event or identifier to remove.
        """
        _change_catalytic_event_list_to_dictlist_after_unpickling(self)

        if isinstance(catalytic_event, str):
            try:
                catalytic_event = self.catalytic_events.get_by_id(catalytic_event)
            except:
                print(
                    f"Catalytic event {catalytic_event} is not related to this enzyme and can thus not be removed!"
                )

        # remove the event from the DictList
        self.catalytic_events.remove(catalytic_event.id)

    def remove_reactions(self, reaction_list: list):
        """Remove reactions from the enzyme variable and remove the reactions from the
        constraint expressions related to the enzyme.

        Args:
            reaction_list (list): A list with Cobra.Reaction objects which should be removed. If a list of identifiers (str)
                is provided, the corresponding enzyme will be obtained from the EnzymeVariables.reaction attribute.
        """
        # check the input
        if not hasattr(reaction_list, "__iter__"):
            enzyme_list = [reaction_list]
        if len(reaction_list) == 0:
            return None

        # check whether the input is an PAModelpy.Package.Enzyme or string and find the corresponding enzyme if needed
        for i, rxn in enumerate(reaction_list):
            if isinstance(rxn, str):
                try:
                    reaction_list[i] = self.reactions.get_by_id(rxn)
                except:
                    print(
                        f"Reaction {rxn} is not associated with the enzyme variable {self.id}. This reaction cannot be removed. \n"
                    )
                    pass

        for rxn in reaction_list:
            # remove from kcat dict
            del self.kcats[rxn.id]
            # remove from reactions dictlist
            self.reaction.remove(rxn)
            # set coefficient in constraint to 0
            for constraint in self.enzyme._constraints.values():
                self.constraints[constraint.name] = constraint
                coeff = 0
                # set coefficients to 0
                if constraint.name[-1] == "f":
                    constraint.set_linear_coefficients(
                        {rxn.forward_variable: coeff, self.forward_variable: 0}
                    )

                elif constraint.name[-1] == "b":
                    constraint.set_linear_coefficients(
                        {rxn.reverse_variable: coeff, self.reverse_variable: 0}
                    )
                # remove constraint from list of r=constraints
                del self.constraints[constraint.name]

    def change_kcat_values(self, reaction_kcat_dict: dict):
        """Changes kcat values for the enzyme variable.

        Args:
            reaction_kcat_dict (dict): A nested dictionary with Cobra.Reaction, kcat key, value pairs to connect the
                enzyme with the associated reaction. The kcat is another dictionary with `f` and `b` for the forward and
                backward reactions, respectively.
        """
        _change_catalytic_event_list_to_dictlist_after_unpickling(self)

        # apply changes to internal dicts (one by one to avoid deleting kcat values)
        kcats_change = {}
        for rxn, kcat_dict in reaction_kcat_dict.items():
            # save change in dict
            self.kcats[rxn.id] = kcat_dict
            for direction, kcat in kcat_dict.items():
                if direction != "f" and direction != "b":
                    warn(
                        f"Invalid direction {direction} for kcat value for enzyme variable {self.id}! Skip!"
                    )
                    continue

                kcats_change[direction] = kcat

            # is enzyme variable already integrated into a model
            if self._model is None:
                warn(f"Catalytic event {self.id} is not integrated into a model!")

            for direction, kcat in kcats_change.items():
                self._model._change_kcat_in_enzyme_constraint(rxn, self.id,
                                                              direction, kcat)
            self._model.solver.update()

    def __str__(self) -> str:
        """Return enzyme variable id as str.

        Returns:
        str
            A string comprised out of the enzyme id
        """
        return f"{self.id}"

    def __copy__(self) -> "PAModelpy.Enzyme.EnzymeVariable":
        """Copy the enzyme variable.

        Returns:
            PAModelpy.Enzyme.EnzymeVariable: A new enzyme variable that is a copy of the original enzyme variable.
        """

        cop = copy(super(EnzymeVariable, self))
        return cop

    def __deepcopy__(self, memo: dict) -> "PAModelpy.Enzyme.EnzymeVariable":
        """Copy the enzyme variable with memo.

        Args:
            memo (dict): Automatically passed parameter.

        Returns:
            PAModelpy.Enzyme.EnzymeVariable: A new enzyme variable that is a copy of the original enzyme variable with memo.
        """

        cop = deepcopy(super(EnzymeVariable, self), memo)
        return cop

    def __getstate__(self):
        # Return the state to be pickled
        state = self.__dict__.copy()
        state['catalytic_events'] = list(self.catalytic_events)

        # Handle any non-serializable attributes here
        return state

    def __setstate__(self, state):
        # Restore state from the unpickled state

        self.__dict__.update(state)
        # Handle any attributes that require initialization or special handling here



from warnings import warn
from copy import copy, deepcopy
from cobra import Object, Gene, Model
from typing import Union

from .Enzyme import Enzyme, EnzymeComplex
from .configuration import Config


class Sector(Object):
    TOTAL_PROTEIN_CONSTRAINT_ID = Config.TOTAL_PROTEIN_CONSTRAINT_ID

    def get_tpc_metabolite(self, model):
        tpc = self.TOTAL_PROTEIN_CONSTRAINT_ID
        try:
            tpc_metabolite = model.constraints[tpc]
            totprot = True
        except:
            tpc_metabolite = None
            totprot = False
        return totprot, tpc_metabolite

    def __copy__(self) -> "Sector":
        """Copy the Sector.

        Returns:
            Sector: A new Sector that is a copy of the original Sector.
        """

        cop = copy(super(Sector, self))
        return cop

    def __deepcopy__(self, memo: dict) -> "Sector":
        """Copy the Sector with memo.

        Args:
            memo (dict): Automatically passed parameter.

        Returns:
            Sector: A new Sector that is a copy of the original Sector with memo.
        """

        cop = deepcopy(super(Sector, self), memo)
        return cop


class EnzymeSector(Sector):
    DEFAULT_MOL_MASS = 3.947778784340140e04  # mean enzymes mass E.coli [g/mol]

    def __init__(self, id_list, mol_mass, configuration=Config):
        self.id_list = id_list  # list with reaction ids which are associated with the specific sector
        self.mol_mass = (
            mol_mass  # dict with molar masses for each enzyme in the id_list (g/mol)
        )
        self.variables = []
        self.constraints = []
        self.model = None
        self.id = ""
        self.slope = 0
        self.intercept = 0

        self.TOTAL_PROTEIN_CONSTRAINT_ID = configuration.TOTAL_PROTEIN_CONSTRAINT_ID


class ActiveEnzymeSector(Sector):
    DEFAULT_MOL_MASS = 3.947778784340140e04  # mean enzymes mass E.coli [g/mol]
    #class with all the information on enzymes related to metabolic reactions
    def __init__(self, rxn2protein: dict, protein2gene:dict = {},configuration:Config =Config):
        """
        _summary_

        Parameters:
            rxn2protein (dict): A dictionary with reaction ID, enzymes_dict key, value pairs for each reaction in the active_enzyme sector.
                The enzymes_dict contains the enzyme identifiers of the enzymes related to the specific reaction as keys, and a dict
                with information about the enzyme as values. The information included in this dict includes the turnover number for
                the forward and backward reaction (1/s), molar mass of the enzyme (mol/g), gene identifiers related to the enzyme,
                and with which other enzymes it forms a complex.
            protein2gene: dict
            enzyme_id, gene_list key, value pairs for each enzyme.The gene_list value is a list of lists which indicates
            'and' or 'or' relationships between the genes which code for the enzyme(complex).

        Example:
            ```
            For the Parameter rxn2protein a dictionary may look like this:
            {
                'R1':
                    {E1:
                                {'f': forward kcat, 'b': backward kcat, 'molmass': molar mass, 'genes': [G1, G2],
                                'protein_reaction_association': [[E1, E2]]
                                },
                            E2:
                                {'f': forward kcat, 'b': backward kcat, 'molmass': molar mass, 'genes': [G3, G4],
                                 'protein_reaction_association': [[E1, E2]]
                    }
            }

            For the Parameter protein2gene a dictionary may look like this:
            {E1:
                    [[gene1], [gene2, gene3]],
            E2:
                    [[gene4]]
            }
            where the gene-protein-reaction associations are the following:
            E1: gene1 or (gene2 and gene3)
            E2: gene4
            ```
        Args:
        rxn2protein : nested dict
            reaction id, enzymes_dict key, value pairs for each reaction in the active_enzyme sector.
             The enzymes_dict contains the enzyme identifiers of the enzymes which are related to the specific reaction
             as keys and a dict with information about the enzyme as values. The information included in this dict is:
              turnover number for forward and backward reaction [1/s], molar mass of the enzyme [mol/g], gene identifiers
              related to the enzyme, with which other enzymes it forms a complex.

        protein2gene: dict
            enzyme_id, gene_list key, value pairs for each enzyme.The gene_list value is a list of lists which indicates
            'and' or 'or' relationships between the genes which code for the enzyme(complex).

        configuration: Config object, optional
                    Information about general configuration of the model including identifier conventions.
                    Default as defined in the `PAModelpy.configuration` script for the E.coli iML1515 model.
        """
        self.TOTAL_PROTEIN_CONSTRAINT_ID = configuration.TOTAL_PROTEIN_CONSTRAINT_ID

        self.rxn2protein = rxn2protein
        self.protein2gene = protein2gene
        # ID of the sector
        self.id = "ActiveEnzymeSector"
        self.model = None

        # for easy removal
        self.variables = []
        self.constraints = []

    def add(self, model):
        self.model = model
        # make and enzymatic constraint for an existing stoichiometric reaction in the form of: e_i=v_i*kcat_i
        # c.f. the GECKO formulation of Sanchez et al. 2017
        print("Add active protein sector\n")
        # check if the constraint exists
        totprot, tpc_metabolite = self.get_tpc_metabolite(model)
        if not totprot:
            warn(
                "Total protein constraint is not in the model. Skip the Active Enzyme Sector"
            )
            return model

        rxn2protein = self.rxn2protein.copy()

        for rxn_id, enzymes in self.rxn2protein.items():
            # extract reaction from model
            if rxn_id not in model.reactions:
                #     reaction = model.reactions.get_by_id(rxn_id)
                # else:
                try:
                    reaction = model.reactions.query(rxn_id)
                    for rxn in reaction:
                        rxn2protein[rxn.id] = enzymes

                        del rxn2protein[rxn_id]
                except:
                    warn(
                        f"Reaction {rxn_id} is not in the model, this reaction will be skipped"
                    )
                    if rxn_id in rxn2protein.keys():
                        del rxn2protein[rxn_id]

        for rxn_id, enzymes in rxn2protein.items():
            reaction = model.reactions.get_by_id(rxn_id)
            # skip the reaction if a problem is encountered in check_kcat_values()
            consistent = self.check_kcat_values(model, reaction, enzymes)
            if not consistent:
                continue

            for enzyme_id, enzyme_dict in enzymes.items():
                kcat = dict(
                    (k, v) for k, v in enzyme_dict.items() if k == "f" or k == "b"
                )
                if 'protein_reaction_association' in enzyme_dict.keys():
                    protein_reaction = enzyme_dict['protein_reaction_association']
                else:
                    protein_reaction = enzyme_id

                # get molar mass of enzyme or replace with default value
                if "molmass" in enzyme_dict.keys():
                    molmass = enzyme_dict["molmass"]
                else:
                    molmass = self.DEFAULT_MOL_MASS

                # check if there already exists an Enzyme object for the EC numbers associated to this reaction,
                # otherwise create one and store all information
                if enzyme_id in model.enzyme_variables and not self._enzyme_is_enzyme_complex(protein_reaction, enzyme_id):
                    enzyme = model.enzymes.get_by_id(enzyme_id)
                    self._add_reaction_to_enzyme(model, enzyme, rxn_id, kcat)
                else:
                    if self.protein2gene != {}:
                        gene_list = self._get_model_genes_from_enzyme(enzyme_id, model)
                    else:
                        gene_list = []

                    enzyme = Enzyme(
                        id=enzyme_id,
                        rxn2kcat={rxn_id: kcat},
                        rxn2pr={rxn_id: protein_reaction},
                        molmass=molmass,
                        genes=gene_list
                    )
                    if self._enzyme_is_enzyme_complex(protein_reaction, enzyme_id):
                        for pr in protein_reaction:
                            if len(pr) > 1:
                                enzyme_complex_id = '_'.join(pr)
                                if enzyme_complex_id not in model.enzymes:
                                    enzyme = EnzymeComplex(
                                        id=enzyme_complex_id,
                                        rxn2kcat={rxn_id: kcat},
                                        molmass=0,
                                        genes=gene_list,
                                        enzymes=[enzyme]
                                    )

                                    model.add_enzymes([enzyme])
                                    self.constraints += [enzyme]
                                    self.variables.append(enzyme.enzyme_variable)
                                else:
                                    enz_complex = model.enzymes.get_by_id(enzyme_complex_id)
                                    self._add_reaction_to_enzyme(model, enz_complex, rxn_id, kcat)
                                    enz_complex.reactions.append(rxn_id)
                                    enz_complex.add_enzymes([enzyme])
                    else:
                        model.add_enzymes([enzyme])

                        self.constraints += [enzyme]
                        self.variables.append(enzyme.enzyme_variable)
                    #in the add enzymes function the enzyme with corresponding catalytic events will be added to the model
                    #and connected to the reaction and total protein constraint


                    # adding to the enzyme sector object for easy removal
                    model.tpc += 1
        return model

    def check_kcat_values(self, model, reaction, enzyme_dict):
        """
        Function to check if the kcat values provided for an enzyme are consistent with the direction of the reaction
        Args:
            model (cobra.Model or PAModel): Model to which the kcat values should be added.
            reaction (cobra.Reaction): Reaction that is catalyzed with the enzyme related to the kcats.
            enzyme_dict (dict): A dictionary with the turnover values for the forward and/or backward reaction for different enzymes [/s].

        Example:
            Example dictionary for the `enzyme_dict` parameter
            ```
            {'E1': {'f': forward kcat, 'b': backward kcat}}
            ```
        """
        if reaction.id not in model.reactions:
            warn(
                "Reaction " + reaction.id + " not in the model. Skip enzyme constraint"
            )
            return False
        # check if there is an EC number associated to each reaction, otherwise, use the reaction id
        if reaction.id not in self.rxn2protein.keys():
            self.rxn2protein[reaction.id] = reaction.id

        # check consistency between provided kcat values and reaction direction
        directions = []
        kcats = []
        for enzyme_info in enzyme_dict.values():
            # get all directions from the kcat dict
            for key, value in enzyme_info.items():
                if key == 'f' or key =='b':
                    directions += [key]
                    kcats += [value]

        rxn_dir = "consistent"
        if reaction.lower_bound >= 0 and reaction.upper_bound > 0:
            # reaction is irreversible in the forward direction

            if not "f" in directions:
                rxn_dir = "inconsistent"

        elif reaction.lower_bound < 0 and reaction.upper_bound <= 0:
            # reaction is irreversible in the backward direction
            if not "b" in directions:
                rxn_dir = "inconsistent"
        else:
            # reaction is reversible
            if not "f" in directions and "b" in directions:
                rxn_dir = "inconsistent"

        if rxn_dir == "inconsistent":
            warn(
                reaction.id
                + ": reaction directionality does not match provided kcat values. Skip reaction"
            )
            return False

        # check kcat values input
        for k in kcats:
            if k < 0:
                warn(
                    [
                        'Turnover number for reaction "',
                        reaction.id,
                        '" is invalid. Skip for active enzyme sector',
                    ]
                )
                return False

        return rxn_dir == "consistent"

    def _add_reaction_to_enzyme(self, model, enzyme:Union[Enzyme, EnzymeComplex], rxn_id:str, kcat:dict):
        # check if catalytic event already exists:
        catal_event_id = f'CE_{rxn_id}'
        if catal_event_id in enzyme.catalytic_events:
            return

        if catal_event_id in model.catalytic_events:
            ce = model.catalytic_events.get_by_id(catal_event_id)
            enzyme.add_catalytic_event(ce, kcat)
        else:
            enzyme.create_catalytic_event(rxn_id=rxn_id, kcats=kcat)
            model.add_catalytic_events([enzyme.catalytic_events.get_by_id(f'CE_{rxn_id}')])

    def _enzyme_is_enzyme_complex(self, protein_reaction, enzyme_id):
        return any([((enzyme_id in pr) & (len(pr) > 1)) for pr in protein_reaction])

    def _get_model_genes_from_enzyme(self, enzyme_id: str, model: Model) -> list:
        """
           Retrieves genes associated with the specified enzyme from the model.

           Args:
               enzyme_id (str): The identifier of the enzyme.
               model (Model): The model containing gene information.

           Returns:
               list: A nested list of gene objects associated with the enzyme.
           """
        if enzyme_id not in self.protein2gene.keys(): return []
        gene_id_list = self.protein2gene[enzyme_id]
        gene_list = []
        for genes_or in gene_id_list:
            genes_and_list = []
            for gene_and in genes_or:
                #check if there is an and relation (then gene and should be a list and not a string)
                if isinstance(gene_and, list):
                    for gene in gene_and:
                        if gene not in model.genes:
                            model.genes.append(Gene(gene))
                else:
                    gene = gene_and
                    if gene not in model.genes:
                        model.genes.append(Gene(gene))
                genes_and_list.append(model.genes.get_by_id(gene))
            gene_list.append(genes_and_list)
        return gene_list

    # def __getstate__(self):
    #     # Return the state to be pickled
    #     state = self.__dict__.copy()
    #     # Handle any non-serializable attributes here
    #     return state

    def __setstate__(self, state):
        # Restore state from the unpickled state
        self.__dict__.update(state)
        # Handle any attributes that require initialization or special handling here

class TransEnzymeSector(EnzymeSector):
    DEFAULT_MOL_MASS = 4.0590394e05  # default E. coli ribosome molar mass [g/mol]
    BIOMASS_RXNID = Config.BIOMASS_REACTION

    # class with all the information on enzymes related to translation/ribosomes (which is linearly related)
    def __init__(
        self,
        tps_0,
        tps_mu,
        id_list=[BIOMASS_RXNID],
        mol_mass=None,
        configuration=Config,
    ):
        super().__init__(id_list, mol_mass, configuration)
        BIOMASS_RXNID = configuration.BIOMASS_REACTION
        if self.mol_mass is None:
            self.mol_mass = [self.DEFAULT_MOL_MASS]
        # id_list only contains the model identifier of the biomass formation reaction
        # mol_mass is the molar mass of a fictional ribosome, if assigned, the computed enzyme concentration resamble
        # ribosome concentrations
        self.tps_0 = tps_0  # amount of protein allocated to the translational sector at zero growth (g_p/g_cdw)
        self.tps_mu = tps_mu  # amount of protein allocated to the translational sector at zero growth (g_p/g_cdw)'
        self.id = "TranslationalProteinSector"

        # *1000 to convert units from g/g_cdw to mg/g_cdw
        self.tps_0_coeff = self.tps_0[0] * 1e3
        self.slope = None
        self.set_slope()
        self.intercept = None
        self.set_intercept()

        # for easy removal
        self.variables = []
        self.constraints = []

    def add(self, model):
        print("Add translational protein sector\n")
        if self.mol_mass is None:
            self.mol_mass = [self.DEFAULT_MOL_MASS]
        self.set_tps_0_coeff()
        return self.add_sector(
            model=model, slope=self.tps_mu[0] * 1e3, intersect=self.tps_0_coeff
        )

    def set_tps_0_coeff(self):
        # *1000 to convert units from g/g_cdw to mg/g_cdw
        self.tps_0_coeff = self.tps_0[0] * 1e3

    def set_slope(self):
        # *1000 to convert units from g/g_cdw to mg/g_cdw
        self.slope = self.tps_mu[0] * 1e3  # / (self.mol_mass[0] * 1e-3)

    def set_intercept(self):
        self.set_tps_0_coeff()
        self.intercept = self.tps_0_coeff


class UnusedEnzymeSector(EnzymeSector):
    DEFAULT_MOL_MASS = 3.947778784340140e04  # mean enzymes mass E.coli [g/mol]

    # class with all the information on 'excess' proteins (linear dependent on substrate uptake rate)
    def __init__(self, ups_0, ups_mu, id_list, mol_mass=None, configuration=Config):
        super().__init__(id_list, mol_mass, configuration)
        if self.mol_mass is None:
            self.mol_mass = [self.DEFAULT_MOL_MASS]
        # id_list only contains the model identifier of the substrate uptake
        # mol_mass is the molar mass of a concentration unit of the unused protein sector
        self.ups_0 = ups_0  # amount of protein allocated to the excess enzyme sector at zero substrate uptake
        if isinstance(ups_mu, list):
            self.ups_mu = ups_mu[0]  # slope of linear relation with growth
        else:
            self.ups_mu = ups_mu
        # *1000 to convert units from g/g_cdw to mg/g_cdw
        self.ups_0_coeff = self.ups_0[0] * 1e3
        self.id = 'UnusedEnzymeSector'

        self.slope = self.ups_mu * 1e3
        self.intercept = self.ups_0_coeff

        # for easy removal
        self.variables = []
        self.constraints = []

    def add_to_model(self, model):
        print("Add unused protein sector\n")
        if self.mol_mass is None:
            self.mol_mass = [self.DEFAULT_MOL_MASS]
        return self.add_sector(
            model=model, slope=self.ups_mu * 1e3, intersect=self.ups_0_coeff
        )


class CustomSector(EnzymeSector):
    DEFAULT_ENZYME_MOL_MASS = 3.947778784340140e04  # mean enzymes mass E.coli [g/mol]

    # class with all the information on a custom protein allocation secotr (linear dependent on a user defined model variable)
    def __init__(
        self, id_list, name, cps_0, cps_s, mol_mass=None, configuration=Config
    ):
        super().__init__(id_list, mol_mass, configuration)
        # id_list containts the name of the reaction on which the custom protein sector is linearly dependent
        self.name = name  # name of the protein sector
        self.lin_rxn_id = id_list[
            0
        ]  # model identifier of the reaction from which the custom protein sector is linearly dependent
        self.cps_0 = cps_0  # intercept of the linear function describing protein allocation of the custom protein sector
        self.cps_s = cps_s  # slope of the linear function describing protein allocation of the custom protein sector
        self.cps_0_coeff = self.cps_0[0] * 1e3
        self.id = f"CustomProteinSector_{self.name}"

        # *1000 to convert units from g/g_cdw to mg/g_cdw
        self.slope = self.cps_s[0] * 1e3
        self.intercept = self.cps_0_coeff

        # for easy removal
        self.variables = []
        self.constraints = []

    def add_to_model(self, model):
        print(f"Add custom protein sector {self.name}\n")
        return self.add_sector(
            model=model, slope=self.cps_mu[0] * 1e3, intersect=self.cps_0_coeff
        )



class Config:
    """
    Object with information about model defaults which are used throughout the package:
    - TOTAL_PROTEIN_CONSTRAINT_ID: str,  `TotalProteinConstraint`
    - P_TOT_DEFAULT: float, 0.258 g_p/g_cdw
    - CO2_EXHANGE_RXNID: str, `EX_co2_e`
    - GLUCOSE_EXCHANGE_RXNID: str, `EX_glc__D_e`
    - BIOMASS_REACTION: str, `BIOMASS_Ec_iML1515_core_75p37M`
    - OXYGEN_UPTAKE_RXNID: str, `EX_o2_e`
    - ACETATE_EXCRETION_RXNID: str, `EX_ac_e`
    - PHYS_RXN_IDS: List of str,  `[BIOMASS_REACTION, GLUCOSE_EXCHANGE_RXNID, ACETATE_EXCRETION_RXNID, CO2_EXHANGE_RXNID, OXYGEN_UPTAKE_RXNID,
                        'PGI', 'G6PDH2r', 'EDA', 'CS', 'ICL', 'PPC', 'ME1', 'ME2']`

    Defaults are configured for the iML1515 E.coli model
    """

    TOTAL_PROTEIN_CONSTRAINT_ID = "TotalProteinConstraint"
    P_TOT_DEFAULT = 0.258  # g_protein/g_cdw
    CO2_EXHANGE_RXNID = "EX_co2_e"
    GLUCOSE_EXCHANGE_RXNID = "EX_glc__D_e"
    BIOMASS_REACTION = "BIOMASS_Ec_iML1515_core_75p37M"
    OXYGEN_UPTAKE_RXNID = "EX_o2_e"
    ACETATE_EXCRETION_RXNID = "EX_ac_e"
    PHYS_RXN_IDS = [
        BIOMASS_REACTION,
        GLUCOSE_EXCHANGE_RXNID,
        ACETATE_EXCRETION_RXNID,
        CO2_EXHANGE_RXNID,
        OXYGEN_UPTAKE_RXNID,
        "PGI",
        "G6PDH2r",
        "EDA",
        "CS",
        "ICL",
        "PPC",
        "ME1",
        "ME2",
    ]

    def reset(self):
        """
        Reset the config object to the standard settings for E.coli iML1515.
        """

        self.TOTAL_PROTEIN_CONSTRAINT_ID = "TotalProteinConstraint"
        self.P_TOT_DEFAULT = 0.258  # g_protein/g_cdw
        self.CO2_EXHANGE_RXNID = "EX_co2_e"
        self.GLUCOSE_EXCHANGE_RXNID = "EX_glc__D_e"
        self.BIOMASS_REACTION = "BIOMASS_Ec_iML1515_core_75p37M"
        self.OXYGEN_UPTAKE_RXNID = "EX_o2_e"
        self.ACETATE_EXCRETION_RXNID = "EX_ac_e"
        self.PHYS_RXN_IDS = [
            self.BIOMASS_REACTION,
            self.GLUCOSE_EXCHANGE_RXNID,
            self.ACETATE_EXCRETION_RXNID,
            self.CO2_EXHANGE_RXNID,
            self.OXYGEN_UPTAKE_RXNID,
            "PGI",
            "G6PDH2r",
            "EDA",
            "CS",
            "ICL",
            "PPC",
            "ME1",
            "ME2",
        ]



import pandas as pd
from tabulate import tabulate
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import os
import cobra
import sys
import numpy as np

from IPython.display import display
from IPython import get_ipython

from .configuration import Config

from typing import Union


class PAMValidator(object):
    RESULT_DIR = os.path.join(os.path.split(os.getcwd())[0], "Results")
    OUT_FILE = os.path.join(RESULT_DIR, "flux_rates_vs_glc.xls")
    GLUCOSE_EXCHANGE_RXNID = Config.GLUCOSE_EXCHANGE_RXNID
    OXYGEN_UPTAKE_RXNID = Config.OXYGEN_UPTAKE_RXNID
    BIOMASS_REACTION = Config.BIOMASS_REACTION
    ACETATE_EXCRETION_RXNID = Config.ACETATE_EXCRETION_RXNID
    CO2_EXHANGE_RXNID = Config.CO2_EXHANGE_RXNID
    PHYS_RXN_IDS = Config.PHYS_RXN_IDS
    MW_GLC = 180.15588  # g/mol
    MW_ACETATE = 59.04
    MW_CO2 = 44.009
    GRADIENT_MAX = 11  # mmol/gdw/h
    GRADIENT_STEP = 0.5  # mmol/gdw/h
    GRADIENT_MIN = 0  # mmol/gdw/h

    def __init__(self, model, phys_file, configuration=Config):
        self.GLUCOSE_EXCHANGE_RXNID = configuration.GLUCOSE_EXCHANGE_RXNID
        self.OXYGEN_UPTAKE_RXNID = configuration.OXYGEN_UPTAKE_RXNID
        self.BIOMASS_REACTION = configuration.BIOMASS_REACTION
        self.ACETATE_EXCRETION_RXNID = configuration.ACETATE_EXCRETION_RXNID
        self.CO2_EXHANGE_RXNID = configuration.CO2_EXHANGE_RXNID
        self.PHYS_RXN_IDS = configuration.PHYS_RXN_IDS

        self.model = model
        self.physiology_data = pd.read_excel(phys_file, sheet_name="Fluxes")
        self.yield_data = pd.read_excel(phys_file, sheet_name="Yields")

        self.results = pd.DataFrame(
            columns=["Y_biomass", "Y_acetate", "Y_co2"] + self.PHYS_RXN_IDS
        )
        self._additional_rxns = []

    @staticmethod
    def check_kernel_type():
        try:
            ipy_str = str(type(get_ipython()))
            if "zmqshell" in ipy_str:
                # jupyter
                return True
            if "terminal" in ipy_str:
                # ipython
                return False
        except:
            # terminal
            return False

    def add_rxn_to_save(self, rxn_list: Union[str, list]):
        if isinstance(rxn_list, str):
            rxn_list = [rxn_list]

        self.PHYS_RXN_IDS += rxn_list
        for rxn in rxn_list:
            self.results[rxn] = 0
        self._additional_rxns.append(rxn_list)

    def validate_literature(self):
        self.parse_data()
        self.run_simulations()
        # for pretty printing: check if the code is ran from a jupyter notebook or not
        jupyter = self.check_kernel_type()

        if jupyter:
            pd.set_option("display.max_rows", None)
            display(self.results)
        else:
            print(tabulate(self.results, headers="keys", tablefmt="psql"))
        self.plot_mu()

    def validate_range(
        self,
        c_uptake_rxn: str = GLUCOSE_EXCHANGE_RXNID,
        pfba: bool = False,
        show: bool = True,
        save: bool = False,
    ):
        # reset result dataframe
        self.results = pd.DataFrame(
            columns=["Y_biomass", "Y_acetate", "Y_co2"] + self.PHYS_RXN_IDS
        )
        try:
            self.parse_data()
        except:
            pass
        self.run_simulations_gradient(c_uptake_rxn=c_uptake_rxn, pfba=pfba)
        jupyter = self.check_kernel_type()

        if jupyter:
            pd.set_option("display.max_rows", None)
            display(self.results)
        else:
            print(tabulate(self.results, headers="keys", tablefmt="psql"))
        if show:
            self.plot_vs_mu(c_uptake_rxn)
            if self._additional_rxns != []:
                self.custom_plot(self._additional_rxns, c_uptake_rxn=c_uptake_rxn)
        if save:
            # make result directory if it not already exists
            if not os.path.exists(self.RESULT_DIR):
                os.mkdir(self.RESULT_DIR)
            self.results.to_excel(self.OUT_FILE, engine="openpyxl")

    def parse_data(self):
        self.physiology_data = pd.melt(
            self.physiology_data,
            id_vars="Reaction_ID",
            var_name="Reference",
            value_vars=[
                "Rijsewijk_2011",
                "Nanchen_2006_1",
                "Nanchen_2006_2",
                "Nanchen_2006_3",
                "Nanchen_2006_4",
            ],
        )

    def run_simulations(self):
        i = 0
        print("Running simulations to validate the model...\n")
        # physiology dataset
        for ref in pd.unique(self.physiology_data.Reference):
            # set glucose uptake rate
            df = self.physiology_data[self.physiology_data.Reference == ref]
            bound = df[df.Reaction_ID == self.GLUCOSE_EXCHANGE_RXNID].value.iloc[0]
            self.set_glc_uptake_bounds(bound)
            self.model.optimize()
            if self.model.solver.status == "optimal":
                self.store_results()
        print("Still validating...\n")
        for ref in pd.unique(self.yield_data.Reference):
            df = self.yield_data[self.yield_data.Reference == ref]
            for glc_flux in pd.unique(df.Glucose_IN):
                self.set_glc_uptake_bounds(glc_flux)
                self.model.optimize()
                if self.model.solver.status == "optimal":
                    self.store_results()
                if i == 2:
                    cobra.core.solution.get_solution(self.model).fluxes.to_csv(
                        os.path.join(
                            os.path.split(os.getcwd())[0], "Results", "test_out_mu.tsv"
                        ),
                        sep="\t",
                    )
                i += 1

    def run_simulations_gradient(
        self, pfba: bool = False, c_uptake_rxn: str = GLUCOSE_EXCHANGE_RXNID
    ):
        # run simulations for increasing glucose consumption rates
        print("Running simulations to validate the model...\n")
        self.model.reset_objective()
        for flux in np.arange(self.GRADIENT_MIN, self.GRADIENT_MAX, self.GRADIENT_STEP):
            self.model.change_reaction_bounds(c_uptake_rxn, -flux, -flux)
            # if glucose is not the substrate turn off glucose uptake reaction
            if c_uptake_rxn != self.GLUCOSE_EXCHANGE_RXNID:
                self.set_glc_uptake_bounds(0)
            if pfba:
                self.model.reset_objective()
            self.model.optimize()
            if self.model.solver.status == "optimal":
                if pfba:
                    self.model.pfba(
                        fraction_of_optimum=0.1, proteins=True, reactions=False
                    )
                self.store_results(c_uptake_rxn=c_uptake_rxn)
            print("Done running simulations\n")

    def run_simulations_glc_o2_gradient(
        self,
        oxygen_gradient: list,
        params_to_save: Union[str, list] = "R_TranslationalProteinSector",
    ):
        """
        Function to run simulations of different oxygen gradients for a range of growth rates.

        This will simulate growth for the entire range of glucose concentrations for each oxygen uptake rate as given by the input.

        Parameters:
            oxygen_gradient (list): List of upper bounds for the oxygen uptake reaction to loop over.
            params_to_save (optional): string or list, which parameter(s) to save for further analysis (default: translational protein sector constraint).

        Returns:
            results (list of dataframes): Saves the growth rate, glucose uptake rate, and the user-defined parameters for each oxygen uptake rate in separate dataframes.
        """

        print("Running simulations of glucose and oxygen gradients...\n")
        result = list()
        if isinstance(params_to_save, str):
            params_to_save = [params_to_save]
        rxns_to_save = [
            self.BIOMASS_REACTION,
            self.GLUCOSE_EXCHANGE_RXNID,
            self.OXYGEN_UPTAKE_RXNID,
        ] + params_to_save
        for o2 in oxygen_gradient:
            df = pd.DataFrame(columns=rxns_to_save)
            self.set_rxn_uptake_bounds(self.OXYGEN_UPTAKE_RXNID, o2)
            print(self.model.reactions.get_by_id(self.OXYGEN_UPTAKE_RXNID).bounds)
            for glc_flux in range(0, self.GRADIENT_MAX, self.GRADIENT_STEP):
                self.set_glc_uptake_bounds(glc_flux / 10)
                self.model.reset_objective()
                self.model.optimize()
                if self.model.solver.status == "optimal":
                    df = self.store_manual_results(rxns_to_save, df)
            result += [df]
        return result

    def run_simulations_ups(
        self,
        ups_gradient: list,
        params_to_save: Union[str, list] = "R_TranslationalProteinSector",
    ):
        """
        Function to run simulations with increasing unused enzyme sectors proportions for a range of growth rates.

        This will simulate growth for the entire range of glucose concentrations for a range of fractions of ups_0 as given by the input.

        Parameters:
            ups_gradient (list): List of upper bounds for the oxygen uptake reaction to loop over.
            params_to_save (optional): string or list, which parameter(s) to save for further analysis (default: translational protein sector constraint).

        Returns:
            results (list of dataframes): Saves the growth rate, glucose uptake rate, and the user-defined parameters for each oxygen uptake rate in separate dataframes.
        """

        print(
            "Running simulations of different unused protein intercepts and glucose gradients...\n"
        )
        result = list()
        if isinstance(params_to_save, str):
            params_to_save = [params_to_save]
        rxns_to_save = [
            self.BIOMASS_REACTION,
            self.GLUCOSE_EXCHANGE_RXNID,
            self.OXYGEN_UPTAKE_RXNID,
        ] + params_to_save
        unused_protein_sector = self.model.sectors.get_by_id("UnusedProteinSector")
        ups_0 = unused_protein_sector.ups_0[0]
        for perc in ups_gradient:
            df = pd.DataFrame(columns=rxns_to_save)
            unused_protein_sector.ups_0[0] = ups_0 * perc
            # self.set_rxn_uptake_bounds(self.OXYGEN_UPTAKE_RXNID, perc)
            for glc_flux in range(0, self.GRADIENT_MAX, self.GRADIENT_STEP):
                self.set_glc_uptake_bounds(glc_flux / 10)
                self.model.reset_objective()
                self.model.optimize()
                if self.model.solver.status == "optimal":
                    df = self.store_manual_results(rxns_to_save, df)
            result += [df]
        return result

    def set_glc_uptake_bounds(self, bound: int):
        rxn_glc_in = self.model.reactions.get_by_id(self.GLUCOSE_EXCHANGE_RXNID)
        # check if model is reversible
        if self.GLUCOSE_EXCHANGE_RXNID[-1] == "b":
            rxn_glc_in._lower_bound, rxn_glc_in._upper_bound = bound, bound
        else:
            rxn_glc_in._lower_bound, rxn_glc_in._upper_bound = -bound, -bound
        rxn_glc_in.update_variable_bounds()

    def set_rxn_uptake_bounds(
        self, rxn_id: str, bound: int, upper: bool = True, lower: bool = False
    ):
        rxn = self.model.reactions.get_by_id(rxn_id)
        if rxn_id[-1] == "b":
            if upper:
                rxn._upper_bound = bound
            if lower:
                rxn._lower_bound = bound
        else:
            if upper:
                rxn._lower_bound = -bound
            if lower:
                rxn._upper_bound = -bound
        rxn.update_variable_bounds()

    def store_results(self, c_uptake_rxn: str = GLUCOSE_EXCHANGE_RXNID):
        result = {}
        c_upt_flux = self.model.reactions.get_by_id(c_uptake_rxn).flux
        # store the relevant flux values
        if c_uptake_rxn != self.GLUCOSE_EXCHANGE_RXNID:
            for i, rxn in enumerate(self.PHYS_RXN_IDS):
                if rxn == self.GLUCOSE_EXCHANGE_RXNID:
                    self.PHYS_RXN_IDS[i] = c_uptake_rxn
                    self.results = pd.DataFrame(columns=self.PHYS_RXN_IDS)
                    # self.results = pd.DataFrame(columns=[ 'Y_biomass', 'Y_acetate', 'Y_co2']+self.PHYS_RXN_IDS)
        for id in self.PHYS_RXN_IDS:
            # getting reaction or variable
            try:
                result[id] = self.model.reactions.get_by_id(id).flux
            except:
                result[id] = self.model.variables[id].primal
            # calculate yields based on glucose for exchange reactions
            if (
                ("EX" in id or "BIOMASS" in id)
                and "glc" not in id
                and "_o2" not in id
                and abs(c_upt_flux) > 0
            ):
                if "BIOMASS" in id:
                    id_ext = "biomass"
                    MW = 1
                if "ac" in id:
                    id_ext = "acetate"
                    MW = 1  # self.MW_ACETATE * 1e-3 #[mmol/mol]
                if "_co2" in id:
                    id_ext = "co2"
                    MW = 1  # self.MW_CO2 * 1e-3 #[mmol/mol]
                yield_id = "Y_" + id_ext
                result[yield_id] = (self.model.reactions.get_by_id(id).flux * MW) / (
                    c_upt_flux * self.MW_GLC * 1e-3
                )
        # appending the results to the dataframe
        result_series = pd.Series(result)
        # sort the series to match the format of the output dataframe
        result_series_sorted = result_series[self.results.columns.to_list()]
        # add the resulting row to the result dataframe
        self.results.loc[len(self.results.index)] = result_series_sorted.to_list()

    def store_manual_results(self, vars_to_save: list, df: pd.DataFrame):
        row = dict()
        for id in vars_to_save:
            try:
                # for variables
                row[id] = self.model.variables[id].primal
            except:
                # for reactions:
                row[id] = self.model.reactions.get_by_id(id).flux
        # appending the results to the dataframe
        df = pd.concat(
            [df, pd.DataFrame({key: pd.Series(value) for key, value in row.items()})],
            ignore_index=True,
        )
        return df

    def plot_mu(self):
        print("Plotting the results against reference values. \n")
        # Initialize figure with subplots
        fig = make_subplots(
            rows=2,
            cols=2,
            specs=[
                [{"type": "scatter"}, {"type": "scatter"}],
                [{"type": "scatter"}, {"type": "scatter"}],
            ],
            subplot_titles=(
                "Growth rate",
                "Biomass yield",
                "CO2 excretion",
                "CO2 yield",
                "Acetate excretion",
                "Acetate yield",
            ),
        )
        result = self.results.__deepcopy__()
        result = result.merge(
            self.yield_data, left_on=self.GLUCOSE_EXCHANGE_RXNID, right_on="Glucose_IN"
        )
        # print(tabulate(result, headers='keys', tablefmt='psql'))
        # growth rate
        fig.add_trace(
            go.Scatter(
                x=result[self.BIOMASS_REACTION],
                y=result["mu"],
                mode="markers",
                showlegend=False,
                marker=dict(color="blue"),
            ),
            row=1,
            col=1,
        )
        fig.add_trace(
            go.Scatter(
                x=[0, 0.5, 1.4],
                y=[0, 0.5, 1.4],
                mode="lines",
                showlegend=False,
                line=dict(color="black", dash="dash", width=1),
            ),
            row=1,
            col=1,
        )
        fig.update_xaxes(title_text="Simulated growth rate [1/h]", row=1, col=1)
        fig.update_yaxes(
            title_text="Experimentally determined growth rate [1/h]", row=1, col=1
        )

        # biomass yield
        fig.add_trace(
            go.Scatter(
                x=result["Y_biomass"],
                y=result["Yield_biomass"],
                mode="markers",
                showlegend=False,
                marker=dict(color="blue"),
            ),
            row=1,
            col=2,
        )
        fig.add_trace(
            go.Scatter(
                x=[0, 0.5, 0.6],
                y=[0, 0.5, 0.6],
                mode="lines",
                showlegend=False,
                line=dict(color="black", dash="dash", width=1),
            ),
            row=1,
            col=2,
        )
        fig.update_xaxes(title_text="Simulated biomass yield [g_x/g_glc]", row=1, col=2)
        fig.update_yaxes(
            title_text="Experimentally determined biomass yield [g_x/g_glc]",
            row=1,
            col=2,
        )

        # CO2 yield
        fig.add_trace(
            go.Scatter(
                x=result["Y_co2"],
                y=result["Yield_co2"],
                mode="markers",
                showlegend=False,
                marker=dict(color="blue"),
            ),
            row=2,
            col=2,
        )
        fig.add_trace(
            go.Scatter(
                x=[0, 15, 30],
                y=[0, 15, 30],
                mode="lines",
                showlegend=False,
                line=dict(color="black", dash="dash", width=1),
            ),
            row=2,
            col=2,
        )
        fig.update_xaxes(title_text="Simulated CO2 yield [mmol_x/g_glc]", row=2, col=2)
        fig.update_yaxes(
            title_text="Experimentally determined CO2 yield [mmol_co2/g_glc]",
            row=2,
            col=2,
        )

        # CO2 out
        fig.add_trace(
            go.Scatter(
                x=result[self.CO2_EXHANGE_RXNID],
                y=result["CO2_OUT"],
                mode="markers",
                showlegend=False,
                marker=dict(color="blue"),
            ),
            row=2,
            col=1,
        )
        fig.add_trace(
            go.Scatter(
                x=[0, 15, 30],
                y=[0, 15, 30],
                mode="lines",
                showlegend=False,
                line=dict(color="black", dash="dash", width=1),
            ),
            row=2,
            col=1,
        )
        fig.update_xaxes(
            title_text="Simulated CO2 exchange rate [mmol/g_cdw/h]", row=2, col=1
        )
        fig.update_yaxes(
            title_text="Experimentally determined CO2 exchange rate [mmol/g_cdw/h]",
            row=2,
            col=1,
        )
        # fig.write_image('Results/Images/mu_co2_plots.png') #depends on the kaleido or orca package

        fig.show()
        # fig.write_image('here.png')

    def plot_vs_mu(self, c_uptake_rxn: str = GLUCOSE_EXCHANGE_RXNID):
        print("Plotting the results against reference values. \n")
        # Initialize figure with subplots
        fig = make_subplots(
            rows=2,
            cols=2,
            specs=[
                [{"type": "scatter"}, {"type": "scatter"}],
                [{"type": "scatter"}, {"type": "scatter"}],
            ],
            subplot_titles=(
                "Glucose consumption",
                "Acetate production",
                "O2 consumption",
                "CO2 production",
            ),
        )

        self.results[c_uptake_rxn] = abs(self.results[c_uptake_rxn])
        self.results[self.OXYGEN_UPTAKE_RXNID] = abs(
            self.results[self.OXYGEN_UPTAKE_RXNID]
        )
        self.yield_data[c_uptake_rxn] = abs(self.yield_data[c_uptake_rxn])
        self.yield_data[self.OXYGEN_UPTAKE_RXNID] = abs(
            self.yield_data[self.OXYGEN_UPTAKE_RXNID]
        )
        self.yield_data[self.BIOMASS_REACTION] = abs(
            self.yield_data[self.BIOMASS_REACTION]
        )

        # glucose consumption
        fig.add_trace(
            go.Scatter(
                x=self.results[c_uptake_rxn],
                y=self.results[self.BIOMASS_REACTION],
                showlegend=False,
                mode="lines",
                marker=dict(color="blue"),
            ),
            row=1,
            col=1,
        )
        fig.add_trace(
            go.Scatter(
                x=self.yield_data[c_uptake_rxn],
                y=self.yield_data[self.BIOMASS_REACTION],
                mode="markers",
                showlegend=False,
                line=dict(color="black", dash="dash", width=1),
            ),
            row=1,
            col=1,
        )
        fig.update_yaxes(title_text="Growth rate [1/h]", row=1, col=1)
        fig.update_xaxes(title_text="Glucose consumption [mmol/gdw/h]", row=1, col=1)

        # acetate excretion
        fig.add_trace(
            go.Scatter(
                x=self.results[c_uptake_rxn],
                y=self.results[self.ACETATE_EXCRETION_RXNID],
                showlegend=False,
                mode="lines",
                marker=dict(color="blue"),
            ),
            row=1,
            col=2,
        )
        fig.add_trace(
            go.Scatter(
                x=self.yield_data[c_uptake_rxn],
                y=self.yield_data[self.ACETATE_EXCRETION_RXNID],
                mode="markers",
                showlegend=False,
                line=dict(color="black", dash="dash", width=1),
            ),
            row=1,
            col=2,
        )
        fig.update_xaxes(title_text="Glucose consumption [mmol/gdw/h]", row=1, col=2)
        fig.update_yaxes(
            title_text="Acetate production rate [mmol/gdw/h]", row=1, col=2
        )

        # O2 uptake
        fig.add_trace(
            go.Scatter(
                x=self.results[c_uptake_rxn],
                y=self.results[self.OXYGEN_UPTAKE_RXNID],
                mode="lines",
                showlegend=False,
                marker=dict(color="blue"),
            ),
            row=2,
            col=1,
        )
        fig.add_trace(
            go.Scatter(
                x=self.yield_data[c_uptake_rxn],
                y=self.yield_data[self.OXYGEN_UPTAKE_RXNID],
                mode="markers",
                showlegend=False,
                line=dict(color="black", dash="dash", width=1),
            ),
            row=2,
            col=1,
        )
        fig.update_xaxes(title_text="Glucose consumption [mmol/gdw/h]", row=2, col=1)
        fig.update_yaxes(title_text="O2 consumption rate [mmol/gdw/h]", row=2, col=1)

        # CO2 excretion
        fig.add_trace(
            go.Scatter(
                x=self.results[c_uptake_rxn],
                y=self.results[self.CO2_EXHANGE_RXNID],
                mode="lines",
                showlegend=False,
                marker=dict(color="blue"),
            ),
            row=2,
            col=2,
        )
        fig.add_trace(
            go.Scatter(
                x=self.yield_data[c_uptake_rxn],
                y=self.yield_data[self.CO2_EXHANGE_RXNID],
                mode="markers",
                showlegend=False,
                line=dict(color="black", dash="dash", width=1),
            ),
            row=2,
            col=2,
        )
        fig.update_xaxes(title_text="Glucose consumption [mmol/gdw/h]", row=2, col=2)
        fig.update_yaxes(title_text="CO2 production rate [mmol/gdw/h]", row=2, col=2)

        fig.show()

    def custom_plot(
        self,
        rxn_ids: list,
        valid_dataframe: pd.DataFrame = None,
        xaxis: str = None,
        c_uptake_rxn: str = GLUCOSE_EXCHANGE_RXNID,
    ):
        """
        Function to plot the results of custom reactions.

        Parameters:
            rxn_ids (list of str): Reaction identifiers of the reactions to be plotted.
            valid_dataframe (pandas.DataFrame, optional): A DataFrame with experimental data to validate the results with.
                The columns should be the same as the rxn_id of the reaction to be plotted and the reaction which should be plotted
                on the x-axis (by default the glucose exchange reaction `EX_glc__D_e_b`). If the DataFrame is not provided,
                only the simulation results will be plotted.
            xaxis (str, optional): The reaction identifier of the reaction which should be plotted on the x-axis (default: `EX_glc__D_e_b`).

        Returns:
            Prints scatter plots of the model simulations vs. experimental data points (if provided).
        """

        if xaxis is None:
            xaxis = c_uptake_rxn
        fig = make_subplots(
            rows=1,
            cols=len(rxn_ids),
            specs=[[{"type": "scatter"}] * len(rxn_ids)],
            subplot_titles=tuple(rxn_id[0] for rxn_id in rxn_ids),
        )
        for i, rxn in enumerate(rxn_ids):
            fig.add_trace(
                go.Scatter(
                    x=self.results[xaxis],
                    y=self.results[rxn[0]],
                    mode="lines",
                    showlegend=False,
                    marker=dict(color="blue"),
                ),
                row=1,
                col=i + 1,
            )
            if valid_dataframe is not None:
                fig.add_trace(
                    go.Scatter(
                        x=valid_dataframe[xaxis],
                        y=valid_dataframe[rxn],
                        mode="markers",
                        showlegend=False,
                        line=dict(color="black", dash="dash", width=1),
                    ),
                    row=1,
                    col=i + 1,
                )
            fig.update_xaxes(title_text=xaxis + " [mmol/gdw/h]", row=1, col=i + 1)
            fig.update_yaxes(title_text=rxn[0] + " [mmol/gdw/h]", row=1, col=i + 1)

            fig.show()



import numpy as np
import pandas as pd
from warnings import warn
from typing import List, Optional, Union, Tuple

import logging
from optlang.symbolics import Zero
from optlang import Variable, interface

from cobra.core import Configuration, Model, Reaction
from cobra.util import solver as sutil
from cobra.util import ProcessPool
from cobra.flux_analysis.parsimonious import add_pfba
from cobra.flux_analysis.loopless import loopless_fva_iter

from PAModelpy.src.PAModelpy.Enzyme import Enzyme, EnzymeVariable

#TODO add test functions!

logger = logging.getLogger(__name__)
configuration = Configuration()

def _init_worker(model: "Model", loopless: bool, sense: str) -> None:
    """Initialize a global model object for multiprocessing.

    Parameters
    ----------
    model: cobra.Model
        The model to operate on.
    loopless: bool
        Whether to use loopless version.
    sense: {"max", "min"}
        Whether to maximise or minimise objective.

    """
    global _model
    global _loopless
    _model = model
    _model.solver.objective.direction = sense
    _loopless = loopless

def _fva_step(variable: Union[Variable, EnzymeVariable, Reaction]) -> Tuple[str, float]:
    """Take a step for calculating FVA.

    Parameters
    ----------
    variables: str
        The variable to test.

    Returns
    -------
    tuple of (str, float)
        The variable ID with the flux value.

    """
    global _model
    global _loopless
    # The previous objective assignment already triggers a reset
    # so directly update coefs here to not trigger redundant resets
    # in the history manager which can take longer than the actual
    # FVA for small models
    if isinstance(variable, Variable):
        model_var = _model.variables[variable.name]
        coeff = {model_var:1}
    else:
        model_var_f = _model.variables[variable.forward_variable.name]
        model_var_r = _model.variables[variable.reverse_variable.name]
        coeff = {model_var_f:1, model_var_r:-1}
    #set the fva objective
    _model.solver.objective.set_linear_coefficients(coeff)

    _model.slim_optimize()

    sutil.check_solver_status(_model.solver.status)
    if _loopless and isinstance(variable, Reaction):
        value = loopless_fva_iter(_model, variable)
    elif _loopless:
        warn('variable is not a reaction, thus cannot perform loopless FVA')
    else:
        value = _model.solver.objective.value

    # handle infeasible case
    if value is None:
        value = float("nan")
        logger.warning(
            f"Could not get a feasible solution for variable {variable}, setting it to NaN. "
            "This is usually due to numerical instability."
        )
    zero_coeff = {key:0 for key in coeff.keys()}
    _model.solver.objective.set_linear_coefficients(
       zero_coeff
    )
    return variable, value

def _get_variables(model, variable_type, variable_list):
    variable_mapping = {
        Enzyme: model.enzyme_variables,
        EnzymeVariable: model.enzyme_variables,
        Reaction: model.reactions,
        }
    variables = {}
    if variable_type is None:
        if variable_list is not None:
            for var_id in variable_list:
                if isinstance(var_id, str):
                    variables = {**variables, **{var_id:model.variables[var_id]}}
                else:
                    variables = {**variables, **{var_id.name: model.variables[var_id.name]}}
        elif variable_type is None:
            variables = model.variables
        else: #TODO does not work for some reason
            mrna_ids = [trans.mrna_variable.name for trans in model.transcripts]
            variables = dict(zip(mrna_ids, model.transcripts))
    else:
        variables = {var.id: var for var in variable_mapping[variable_type]}
        if variable_list is not None:
            vars = variables.copy()
            variables = {}
            for var_id, var in vars.items():
                if var_id in variable_list:
                    variables[var_id] = var
            # variables = {var: var.id for var in variable_mapping[variable_type]}
    return variables

def flux_variability_analysis(
    model: Model,
    variable_type: Optional[Union[Enzyme, EnzymeVariable, Reaction]] = None,
    variable_list: Optional[List[Union[Enzyme, EnzymeVariable, Reaction, str]]] = None,
    loopless: bool = False,
    fraction_of_optimum: float = 1.0,
    pfba_factor: Optional[float] = None,
    processes: Optional[int] = None,
) -> pd.DataFrame:
    """Determine the minimum and maximum flux value for each reaction.

    Adjusted from cobra.flux_analysis.variability.flux_variability_analysis

    Args:
    model : cobra.Model
        The model for which to run the analysis. It will *not* be modified.
    variable_type: type of variable to perform FVA on
        if None, it will use all model variables
    variable_list : list of specified variables defined in variably_type or str, optional
        for which to obtain min/max fluxes. If None will use
        all of the specified variables in the model (default None).
    loopless : bool, optional
        Whether to return only loopless solutions. This is significantly
        slower. Please also refer to the notes (default False).
    fraction_of_optimum : float, optional
        Must be <= 1.0. Requires that the objective value is at least the
        fraction times maximum objective value. A value of 0.85 for instance
        means that the objective has to be at least at 85% percent of its
        maximum (default 1.0).
    pfba_factor : float, optional
        Add an additional constraint to the model that requires the total sum
        of absolute fluxes must not be larger than this value times the
        smallest possible sum of absolute fluxes, i.e., by setting the value
        to 1.1 the total sum of absolute fluxes must not be more than
        10% larger than the pFBA solution. Since the pFBA solution is the
        one that optimally minimizes the total flux sum, the `pfba_factor`
        should, if set, be larger than one. Setting this value may lead to
        more realistic predictions of the effective flux bounds
        (default None).
    processes : int, optional
        The number of parallel processes to run. If not explicitly passed,
        will be set from the global configuration singleton (default None).

    Returns
    pandas.DataFrame
        A data frame with variable identifiers as the index and two columns:
        - maximum: indicating the highest possible flux
        - minimum: indicating the lowest possible flux

    Notes:
        - This implements the fast version as described in [1]_. Please note that
        the flux distribution containing all minimal/maximal fluxes does not have
        to be a feasible solution for the model. Fluxes are minimized/maximized
        individually and a single minimal flux might require all others to be
        sub-optimal.

        - Using the loopless option will lead to a significant increase in
        computation time (about a factor of 100 for large models). However, the
        algorithm used here (see [2]_) is still more than 1000x faster than the
        'naive' version using `add_loopless(model)`. Also note that if you have
        included constraints that force a loop (for instance by setting all fluxes
        in a loop to be non-zero) this loop will be included in the solution.

    References:
    .. [1] Computationally efficient flux variability analysis.
       Gudmundsson S, Thiele I.
       BMC Bioinformatics. 2010 Sep 29;11:489.
       doi: 10.1186/1471-2105-11-489, PMID: 20920235

    .. [2] CycleFreeFlux: efficient removal of thermodynamically infeasible
       loops from flux distributions.
       Desouki AA, Jarre F, Gelius-Dietrich G, Lercher MJ.
       Bioinformatics. 2015 Jul 1;31(13):2159-65.
       doi: 10.1093/bioinformatics/btv096.

    """
    variables = _get_variables(model, variable_type, variable_list)
    if processes is None:
        processes = configuration.processes

    num_variables = len(variables)
    processes = min(processes, num_variables)

    fva_result = pd.DataFrame(
        {
            "minimum": np.zeros(num_variables, dtype=float),
            "maximum": np.zeros(num_variables, dtype=float),
        },
        index=variables,
    )
    prob = model.problem
    with model:
        # Safety check before setting up FVA.
        model.slim_optimize(
            error_value=None,
            message="There is no optimal solution for the chosen objective!",
        )
        # Add the previous objective as a variable to the model then set it to
        # zero. This also uses the fraction to create the lower/upper bound for
        # the old objective.
        if model.solver.objective.direction == "max":
            fva_old_objective = prob.Variable(
                "fva_old_objective",
                lb=fraction_of_optimum * model.solver.objective.value,
            )
        else:
            fva_old_objective = prob.Variable(
                "fva_old_objective",
                ub=fraction_of_optimum * model.solver.objective.value,
            )
        fva_old_obj_constraint = prob.Constraint(
            model.solver.objective.expression - fva_old_objective,
            lb=0,
            ub=0,
            name="fva_old_objective_constraint",
        )
        model.add_cons_vars([fva_old_objective, fva_old_obj_constraint])

        if pfba_factor is not None:
            if pfba_factor < 1.0:
                warn(
                    "The 'pfba_factor' should be larger or equal to 1.",
                    UserWarning,
                )
            with model:
                add_pfba(model, fraction_of_optimum=0)
                ub = model.slim_optimize(error_value=None)
                flux_sum = prob.Variable("flux_sum", ub=pfba_factor * ub)
                flux_sum_constraint = prob.Constraint(
                    model.solver.objective.expression - flux_sum,
                    lb=0,
                    ub=0,
                    name="flux_sum_constraint",
                )
            model.add_cons_vars([flux_sum, flux_sum_constraint])

        model.objective = Zero  # This will trigger the reset as well
        for what in ("minimum", "maximum"):
            if processes > 1:
                # We create and destroy a new pool here in order to set the
                # objective direction for all reactions. This creates a
                # slight overhead but seems the most clean.
                chunk_size = len(variables) // processes
                with ProcessPool(
                    processes,
                    initializer=_init_worker,
                    initargs=(model, loopless, what[:3]),
                ) as pool:
                    for var, value in pool.imap_unordered(
                        _fva_step, variables.values(), chunksize=chunk_size
                    ):
                        try: var_id = var.id
                        except: var_id = var.name
                        fva_result.at[var_id, what] = value
            else:
                _init_worker(model, loopless, what[:3])
                for var, value in map(_fva_step, variables):
                    try: var_id = var.id
                    except: var_id = var.name
                    fva_result.at[var_id, what] = value

    return fva_result[["minimum", "maximum"]]


"""
CatalyticEvent object which relates Reaction variables to the EnzymeVariable and Enzyme objects.
It contains multiple functions which enable easy mapping and handling of one Event of catalysis
(e.g. one conversion of substrate to product, can be catalyzed by multiple enzymes)
"""
import cobra
from cobra import DictList, Object, Reaction
from cobra.exceptions import OptimizationError
from cobra.util.solver import check_solver_status
from optlang.symbolics import Zero
from typing import Optional, Dict, Union
from warnings import warn
from copy import copy, deepcopy


class CatalyticEvent(Object):
    """
   CatalyticEvent is a class for holding information regarding the
        catalysis of a Reaction in a cobra.Model object. It serves as an interface
        between the metabolic reaction and the associated enzyme constraints and variables.

    Notes:
        There are three different scenarios:
        - Enzyme complex: multiple enzymes together are associated with an EnzymeComplex object
        - Isozymes: multiple enzymes independently associated with a single catalytic event
        - Other: a single enzyme is associated with a single catalytic event

    Parameters:
        kcats2enzymes (dict): A dictionary with enzyme, kcat key, value pairs to connect the enzyme with the associated reaction.
            The kcat is another dictionary with 'f' and 'b' for the forward and backward reactions, respectively.
        id (str, optional): The identifier to associate with this catalytic event (default None).
        rxn_id (str, optional): The reaction with which this catalytic event is associated.
        name (str, optional): A human-readable name for the reaction (default "").
    """

    def __init__(
        self,
        kcats2enzymes: Dict,
        id: Optional[str] = None,  # ID of enzymatic reaction
        rxn_id: str = "",  # ID of reaction
        name: str = "",
    ):
        # identification
        self.id = id
        self.name = name

        # relation to reaction
        self.rxn_id = rxn_id
        self.rxn = None
        self.catalytic_reactions = DictList()

        # relation to enzymes
        self.kcats = kcats2enzymes
        self.enzymes = DictList()
        for enzyme in kcats2enzymes.keys():
            self.enzymes.append(enzyme)
        self.enzyme_variables = DictList()

        # other attributes
        self.constraints = {} # store IDs of constraint the catalytic event is associated with
        self.variables = dict()
        self._model = None
        self.annotation = {"type": "Constraint"}

    @property
    def kcat_values(self):
        """returns a dictionary with kcat values and enzymes"""
        return self.kcats

    @property
    def flux(self) -> float:
        """
        Get the flux value in the most recent solution.

        Flux is the primal value of the corresponding variable in the model.

        Returns:
            flux (float): Flux is the primal value of the corresponding variable in the model.

        Warnings:
            * Accessing reaction fluxes through a `Solution` object is the safer,
              preferred, and only guaranteed to be correct way. You can see how to
              do so easily in the examples.
            * Reaction flux is retrieved from the currently defined
              `self._model.solver`. The solver status is checked but there are no
              guarantees that the current solver state is the one you are looking
              for.
            * If you modify the underlying model after an optimization, you will
              retrieve the old optimization values.

        Raises:
            RuntimeError: If the underlying model was never optimized beforehand or the
                reaction is not part of a model.
            OptimizationError: If the solver status is anything other than `optimal`.
            AssertionError: If the flux value is not within the bounds.

        Examples:
            ```
            >>> from cobra.io import load_model
            >>> model = load_model("textbook")
            >>> solution = model.optimize()
            >>> model.variables.PFK.flux
            7.477381962160283
            >>> solution.fluxes.PFK
            7.4773819621602833
            ```
        """

        try:
            check_solver_status(self._model.solver.status)
            total_flux = 0
            for enzyme_variable in self.enzyme_variables:
                fwd_flux = enzyme_variable.variables["forward"].primal
                rev_flux = enzyme_variable.variables["reverse"].primal
                total_flux += fwd_flux + rev_flux
            return total_flux
        except AttributeError:
            raise RuntimeError(f"reaction '{self.id}' is not part of a model")
        # Due to below all-catch, which sucks, need to reraise these.
        except (RuntimeError, OptimizationError) as err:
            raise err
        # Would love to catch CplexSolverError and GurobiError here.
        except Exception as err:
            raise OptimizationError(
                f"Likely no solution exists. Original solver message: {str(err)}."
            ) from err

    @property
    def concentration(self) -> float:
        """
        Get the enzyme concentration value of the most recent solution.
        The enzyme concentration equals the flux value.

        Returns:
            float: Enzyme concentration in [mmol/gDW].
        """

        return self.flux

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model):
        self._model = model

        # add reaction instance
        if self.rxn_id in self._model.reactions:
            self.rxn = self._model.reactions.get_by_id(self.rxn_id)
        else:  # create new reaction and add to model
            rxn = cobra.Reaction(id=self.rxn_id)
            self._model.add_reactions([rxn])
            self.rxn = rxn
        # add reaction constraint
        ce_constraint = self._model.problem.Constraint(Zero, name= self.id, lb = 0, ub=0)
        self._model.add_cons_vars([ce_constraint])
        self._model.constraints[self.id].set_linear_coefficients({
            self.rxn.forward_variable: 1,
            self.rxn.reverse_variable: -1,
        })
        self.constraints[self.id] = ce_constraint

        #add enzymes to the model if they are not there already
        for enzyme in self.enzymes:
            if enzyme in self._model.enzymes:
                self.constraints = {**self.constraints, **enzyme._constraints}
                enzyme._constraints[self.id] = ce_constraint
                enzyme_model = self._model.enzymes.get_by_id(enzyme.id)

                if self.rxn_id in enzyme_model.rxn2kcat.keys():
                    if self not in enzyme_model.catalytic_events:
                        enzyme_model.add_catalytic_event(
                            self, kcats={enzyme: enzyme.rxn2kcat[self.rxn_id]}
                        )
                else:
                    print(
                        f"Reaction {self.rxn_id} is not related to enzyme {enzyme.id}"
                    )
            else:
                self._model.add_enzymes([enzyme])

            # add enzyme-reaction association
            self.add_enzyme_reaction_association(enzyme)



    def add_enzymes(self, enzyme_kcat_dict: dict):
        """
        Add enzymes to the catalytic event and create bindings to the related model.
        The enzymes in the enzyme_kcat_dict are individual isozymes. Enzyme complexes
        should be added as an EnzymeComplex object with a single kcat value.

        Parameters:
            enzyme_kcat_dict: Dict
                A nested dictionary with enzyme, kcat key, value pairs to connect the
                enzyme with the associated reaction. The kcat is another dictionary with `f` and `b`
                for the forward and backward reactions respectively.
        """
        # return lists back to dictlist after unpickling
        if isinstance(self.enzymes, list):
            self.enzymes = DictList(self.enzymes)
            self.enzyme_variables = DictList(self.enzyme_variables)

        for enzyme, kcat in enzyme_kcat_dict.items():
            # check if the enzyme is already associated to the catalytic event
            if enzyme in self.enzymes:
                self.add_enzyme_reaction_association(enzyme)
                self.change_kcat_values(
                    {enzyme.id: enzyme.get_kcat_values(rxn_ids=[self.rxn_id+"_"+enzyme.id])})
                continue

            self.enzymes.append(enzyme)
            self.kcats[enzyme] = kcat

            if self._model is None:
                continue
            # check if enzyme is in the model
            try:
                self._model.enzymes.get_by_id(enzyme.id)
            # if not: add the enzyme to the model
            except:
                self._model.add_enzymes([enzyme])

            for direction in kcat.keys():
                if direction != "f" and direction != "b":
                    warn(
                        f"Invalid direction {direction} for kcat value for enzyme variable {self.id}! Skip!"
                    )
                    continue

            # add enzyme to catalytic event and the related variable
            # get enzyme variable
            enzyme_var = self._model.enzyme_variables.get_by_id(enzyme.id)
            self.enzyme_variables.append(enzyme_var)

            # add constraints to the catalytic event
            self.constraints = {**self.constraints, **enzyme._constraints}

            # add reaction-protein association
            self.add_enzyme_reaction_association(enzyme)

            #connect the enzyme variable to the enzyme in the model and the reaction
            for direction, kcatvalue in kcat.items():
                #add enzyme to the associated reaction with kinetic constants
                #and relate enzyme to the catalytic event
                if direction == 'f':
                    self.constraints[f'EC_{enzyme.id}_{direction}'].set_linear_coefficients({
                        enzyme_var.forward_variable: -1
                        })

                elif direction == 'b':
                    self.constraints[f'EC_{enzyme.id}_{direction}'].set_linear_coefficients({
                        enzyme_var.reverse_variable: -1
                        })

    def add_enzyme_reaction_association(self, enzyme):
        catalytic_reaction = Reaction(self.id + "_" + enzyme.id, lower_bound=-1e3, upper_bound=1e3)
        self._model.add_reactions([catalytic_reaction])
        self.catalytic_reactions.append(catalytic_reaction)
        self._model.constraints[self.id].set_linear_coefficients({
            catalytic_reaction.forward_variable: -1,
            catalytic_reaction.reverse_variable: 1,
        })
        self.add_catalytic_reaction_to_enzyme_constraint(catalytic_reaction, enzyme)

    def add_catalytic_reaction_to_enzyme_constraint(self, catalytic_reaction:Reaction,
                                                    enzyme):
        #remove relation to metabolic reaction
        for dir in ['f', 'b']:
            self._model._change_kcat_in_enzyme_constraint(self.rxn, enzyme.id,
                                                      dir, 0)
        kcat_dict = self.kcats[enzyme]
        for direction, kcatvalue in kcat_dict.items():
            self._model._change_kcat_in_enzyme_constraint(catalytic_reaction, enzyme.id,
                                                          direction, kcatvalue)

        #change rxn2kcat dict for correct referencing
        del enzyme.rxn2kcat[self.rxn_id]
        enzyme.rxn2kcat[catalytic_reaction.id] = kcat_dict


    def remove_enzymes(self, enzyme_list: list):
        """
        Remove enzymes from the catalytic event and remove the catalytic event from the
        constraint expressions related to the enzyme.

        Parameters:
            enzyme_list: List[Union[str, PAModelpy.Package.Enzyme]]
                A list with PAModelpy.Package.Enzyme objects to be removed. If a list of identifiers (str)
                is provided, the corresponding enzyme will be obtained from the CatalyticEvent.enzymes attribute.
        """
        # return lists back to dictlist after unpickling
        if isinstance(self.enzymes, list):
            self.enzymes = DictList(self.enzymes)
            self.enzyme_variables = DictList(self.enzyme_variables)

        # check the input
        if not hasattr(enzyme_list, "__iter__"):
            enzyme_list = [enzyme_list]
        if len(enzyme_list) == 0:
            return None

        # check wether the input is an PAModelpy.Package.Enzyme or string and find the corresponding enzyme if needed
        for i, enz in enumerate(enzyme_list):
            if isinstance(enz, str):
                try:
                    enzyme_list[i] = self.enzymes.get_by_id(enz)
                except:
                    print(
                        f"Enzyme {enz} is not associated with the catalytic event {self.id}. This enzyme will not be removed. \n"
                    )
                    pass

        for enz in enzyme_list:
            # remove from kcat dict
            del self.kcats[enz]
            # remove from enzymes dictlist
            self.enzymes.remove(enz)
            # remove enzyme variable from dictlist if it still exists
            if self.enzyme_variables.has_id(enz.id):
                enzyme_var = self.enzyme_variables.get_by_id(enz.id)
                self.enzyme_variables.remove(enzyme_var)
            # set coefficient in constraint to 0
            for constraint in set(
                [cons for name, cons in self.constraints.items() if enz.id in name]
            ):
                if constraint in self._model.constraints.values():
                    self._model.remove_cons_vars([constraint])
                # remove constraint from list of r=constraints
                del self.constraints[constraint.name]

            # if there are no enzymes associated to the reaction anymore, the reaction flux will be 0
            if len(self.enzymes) == 0:
                no_enz_constraint_f = self._model.problem.Constraint(
                    Zero, name=f"{self.rxn_id}_no_enzyme_f", lb=0, ub=0
                )
                no_enz_constraint_b = self._model.problem.Constraint(
                    Zero, name=f"{self.rxn_id}_no_enzyme_b", lb=0, ub=0
                )
                self._model.add_cons_vars([no_enz_constraint_f, no_enz_constraint_b])

                self._model.constraints[
                    no_enz_constraint_f.name
                ].set_linear_coefficients({self.rxn.forward_variable: 1})
                self.constraints[no_enz_constraint_f.name] = no_enz_constraint_f

                self._model.constraints[
                    no_enz_constraint_b.name
                ].set_linear_coefficients({self.rxn.forward_variable: 1})
                self.constraints[no_enz_constraint_b.name] = no_enz_constraint_b

    def change_kcat_values(self, enzyme_kcat_dict : dict):
        """changes kcat values for the enzyme variable

        Args:
        enzyme_kcat_dict: nested Dict
            A Dict with enzyme_id, kcat key, value pairs to connect the
            enzyme with the associated reaction the kcat is another dict with 'f' and 'b'
            for the forward and backward reactions respectively.
        
        """

        # apply changes to internal dicts (one by one to avoid deleting kcat values)
        kcats_change = {}
        for enzyme, kcat_dict in enzyme_kcat_dict.items():
            # save change in dict
            self.kcats[enzyme] = kcat_dict
            for direction, kcat in kcat_dict.items():
                if direction != "f" and direction != "b":
                    warn(
                        f"Invalid direction {direction} for kcat value for enzyme variable {self.id}! Skip!"
                    )
                    continue

                kcats_change[direction] = kcat

            # is enzyme variable already integrated into a model
            if self._model is None:
                # warn(f'Catalytic event {self.id} is not integrated into a model!')
                return

            enzyme_var = self._model.enzyme_variables.get_by_id(enzyme)
            if enzyme_var not in self.enzyme_variables:
                self.enzyme_variables.add(enzyme_var)
            ce_reaction = self._model.reactions.get_by_id('CE_'+self.rxn_id+"_"+enzyme)
            enzyme_var.change_kcat_values({ce_reaction: kcat_dict})
            for direction, kcat in kcats_change.items():
                #change enzyme variable
                enzyme_var.kcats[ce_reaction.id][direction] = kcat
                self._model._change_kcat_in_enzyme_constraint(ce_reaction, enzyme, direction, kcat)

    def __copy__(self) -> "CatalyticEvent":
        """
        Copy the CatalyticEvent.

        Returns:
            CatalyticEvent:
                A new CatalyticEvent that is a copy of the original CatalyticEvent.
        """

        cop = copy(super(CatalyticEvent, self))
        return cop

    def __deepcopy__(self, memo: dict) -> "CatalyticEvent":
        """
        Copy the CatalyticEvent with memo.

        Parameters:
            memo (dict): Automatically passed parameter.

        Returns:
            CatalyticEvent:
                A new CatalyticEvent that is a copy of the original CatalyticEvent with memo.
        """

        cop = deepcopy(super(CatalyticEvent, self), memo)
        return cop

    def __getstate__(self):
        # Return the state to be pickled
        state = self.__dict__.copy()
        # Handling non-serializable attributes
        state['enzyme_variables'] = list(self.enzyme_variables)
        state['enzymes'] = list(self.enzymes)
        return state

    def __setstate__(self, state):
        # Restore state from the unpickled state
        self.__dict__.update(state)






from src.PAModelpy import PAModel, Enzyme, EnzymeVariable

# aminoacid lookup table
aa_lookup = {'V': 'VAL', 'I': 'ILE', 'L': 'LEU', 'E': 'GLU', 'Q': 'GLN', \
             'D': 'ASP', 'N': 'ASN', 'H': 'HIS', 'W': 'TRP', 'F': 'PHE', 'Y': 'TYR', \
             'R': 'ARG', 'K': 'LYS', 'S': 'SER', 'T': 'THR', 'M': 'MET', 'A': 'ALA', \
             'G': 'GLY', 'P': 'PRO', 'C': 'CYS'}


def check_freq(x, total):
    return {c: x.count(c) / total for c in set(x)}

def read_sequence_from_file(seq_file_path: str)-> str:
    """ Reads the aminoacid composition from a ...file

    Args:
        seq_file_path: path to the file containing the sequence information
    Returns:
        aa_seq: aminoacid sequence (1 Letter code) as string
    """
    # read amino acid sequence
    with open(seq_file_path) as f:
        lines = f.readlines()
    # need to remove document start ('\ufeff') and end ('\n') to only yield the amino acid sequence
    aa_seq = lines[0].strip().replace('\ufeff', '')
    return aa_seq

def match_aminoacid_to_model_identifiers_and_frequency(aa_seq:str) -> dict:
    """ Calculates the frequency of occurance of an aminoacid in a sequence
    Also maps the aminoacids 1 letter codes to the corresponding 3 letter code
    and the BiGG identifier

    Args:
        aa_seq: aminoacid sequence parsed as string
    Returns:
        aa_biggid_freq: dict with bigg_id: a_frequency key: value pairs. The bigg_id represents the
        identifier related to the metabolite related to the specific aminoacid
    """

    # determine amino acid composition
    aa_freq = check_freq(aa_seq, len(aa_seq))

    # match to model identifiers
    aa_biggid_freq = dict()
    for aa, freq in aa_freq.items():
        threeletter = aa_lookup[aa].lower()
        if threeletter != 'gly':
            bigg_id = f'{threeletter}__L_c'
        else:
            bigg_id = f'{threeletter}_c'
        aa_biggid_freq[bigg_id] = freq
    return aa_biggid_freq

def add_aminoacid_sequence(model: PAModel, seq:dict, protein:EnzymeVariable):
    """
    model: COBRA model
    seq: dict with {aminoacid_id: freq} key, value pairs
    protein: enzyme variable
    """
    for aa, freq in seq.items():
        if aa not in model.constraints.keys(): continue
        model.constraints[aa].set_linear_coefficients({
            protein.forward_variable: -freq / protein.molmass,
            protein.reverse_variable: -freq / protein.molmass
        })
    return model

def add_recombinant_protein_to_pam(pam:PAModel, protein:Enzyme, aa_seq: str) -> PAModel:
    pam.add_enzymes(protein)
    aa_to_freq = match_aminoacid_to_model_identifiers_and_frequency(aa_seq.replace(' ', ''))
    add_aminoacid_sequence(pam, aa_to_freq, pam.enzyme_variables.get_by_id(protein.id))
    return pam